"""Functions related to the plotting of the e-models."""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import copy
import glob
import logging
import re
from pathlib import Path

import efel
import matplotlib
import matplotlib.font_manager
import matplotlib.pyplot as plt
import numpy
from bluepyopt.ephys.locations import NrnSeclistCompLocation
from bluepyopt.ephys.protocols import SweepProtocol
from bluepyopt.ephys.recordings import CompRecording
from bluepyopt.ephys.stimuli import NrnSquarePulse
from matplotlib import cm
from matplotlib import colors

from bluepyemodel.data.utils import read_dendritic_data
from bluepyemodel.evaluation.evaluation import compute_responses
from bluepyemodel.evaluation.evaluation import get_evaluator_from_access_point
from bluepyemodel.evaluation.evaluator import PRE_PROTOCOLS
from bluepyemodel.evaluation.evaluator import add_recordings_to_evaluator
from bluepyemodel.evaluation.protocols import ThresholdBasedProtocol
from bluepyemodel.evaluation.utils import define_bAP_feature
from bluepyemodel.evaluation.utils import define_bAP_protocol
from bluepyemodel.evaluation.utils import define_EPSP_feature
from bluepyemodel.evaluation.utils import define_EPSP_protocol
from bluepyemodel.model.morphology_utils import get_basal_and_apical_max_radial_distances
from bluepyemodel.tools.utils import make_dir
from bluepyemodel.tools.utils import read_checkpoint
from bluepyemodel.tools.utils import select_rec_for_thumbnail

# pylint: disable=W0612,W0102,C0209

matplotlib.rcParams["pdf.fonttype"] = 42

logger = logging.getLogger("__main__")
logging.getLogger("matplotlib").setLevel(level=logging.ERROR)

colours = {
    "datapoint": "orangered",
    "dataline": "red",
    "modelpoint_apical": "cornflowerblue",
    "modelline_apical": "darkblue",
    "modelpoint_basal": "mediumseagreen",
    "modelline_basal": "darkgreen",
}


def save_fig(figures_dir, figure_name, dpi=100):
    """Save a matplotlib figure"""
    p = Path(figures_dir) / figure_name
    plt.savefig(str(p), dpi=dpi, bbox_inches="tight")
    plt.close("all")
    plt.clf()


def get_traces_ylabel(var):
    """Get ylabel for traces subplot."""
    if var == "v":
        return "Voltage (mV)"
    if var[0] == "i":
        return "Current (pA)"
    if var[-1] == "i":
        return "Ionic concentration (mM)"
    return ""


def get_recording_names(protocol_config, stimuli):
    """Get recording names which traces are to be plotted.

    Does not return extra ion / current recordings.

    Args:
        protocol_config (list): list of ProtocolConfiguration from FitnessCalculatorConfiguration
        stimuli (list): list of all protocols (protocols from configuration + pre-protocols)
    """
    # recordings from fitness calculator
    recording_names = {
        recording["name"] for prot in protocol_config for recording in prot.recordings_from_config
    }

    # expects recording names to have prot_name.location_name.variable structure
    prot_names = {rec_name.split(".")[0] for rec_name in recording_names}

    # add pre-protocol recordings
    # expects pre-protocol to only have 1 recording
    pre_prot_rec_names = {
        protocol.recordings[0].name
        for protocol in stimuli.values()
        if protocol.name not in prot_names and protocol.recordings
    }
    recording_names.update(pre_prot_rec_names)

    return recording_names


def get_traces_names_and_float_responses(responses, recording_names):
    """Extract the names of the traces to be plotted, as well as the float responses values."""

    traces_names = []
    threshold = None
    holding = None
    rmp = None
    rin = None

    for resp_name, response in responses.items():
        if not (isinstance(response, float)):
            if resp_name in recording_names:
                traces_names.append(resp_name)

        else:
            if resp_name == "bpo_threshold_current":
                threshold = response
            elif resp_name == "bpo_holding_current":
                holding = response
            elif resp_name == "bpo_rmp":
                rmp = response
            elif resp_name == "bpo_rin":
                rin = response

    return traces_names, threshold, holding, rmp, rin


def get_title(emodel, iteration, seed):
    """Returns 'emodel ; iteration={iteration} ; seed={seed}'

    Args:
        emodel (str): emodel name
        iteration (str): githash
        seed (int): random number seed
    """
    title = str(emodel)
    if iteration is not None:
        title += f" ; iteration = {iteration}"
    if seed is not None:
        title += f" ; seed = {seed}"
    return title


def optimisation(
    optimiser,
    emodel,
    iteration,
    seed,
    checkpoint_path="./checkpoint.pkl",
    figures_dir="./figures",
    write_fig=True,
):
    """Create plots related to a BluePyOpt optimisation"""

    make_dir(figures_dir)
    run, _ = read_checkpoint(checkpoint_path)

    ngen = run["logbook"].select("gen")
    is_finished_msg = ""
    if "CMA" in optimiser:
        is_finished_msg = f"is finished: {not run['CMA_es'].active}"

    legend_text = "\n".join(
        (
            f"min score = {min(run['logbook'].select('min')):.3f}",
            f"# of generations = {run['generation']}",
            f"evolution algorithm: {optimiser}",
            is_finished_msg,
        )
    )

    fig, axs = plt.subplots(1, figsize=(8, 8), squeeze=False)

    title = get_title(emodel, iteration, seed)
    axs[0, 0].set_title(title)

    axs[0, 0].plot(ngen, run["logbook"].select("min"), label="Minimum", c="black")

    axs[0, 0].plot(ngen, run["logbook"].select("avg"), label="Average", c="grey")

    axs[0, 0].set_yscale("log")
    axs[0, 0].set_xlabel("Number of generations")
    axs[0, 0].set_ylabel("Fitness")
    axs[0, 0].legend(title=legend_text, loc="upper right", frameon=False)

    p = Path(checkpoint_path)

    figure_name = p.stem
    figure_name += "__optimisation.pdf"

    plt.tight_layout()

    if write_fig:
        save_fig(figures_dir, figure_name)

    return fig, axs


def _create_figure_parameter_histograms(
    histograms,
    evaluator,
    metadata,
    seed,
    max_n_gen,
    gen_per_bin,
    figures_dir,
    write_fig,
):
    """Create figure and plot the data for the evolution of the density of parameters."""

    ncols = 5
    nrows = len(evaluator.params) // ncols + 1

    fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=(25, 2 * nrows))
    axs = axs.flat

    # Plot the histograms
    for param_index, param in enumerate(evaluator.params):
        axs[param_index].imshow(
            100.0 * numpy.flip(histograms[param_index].T, 0),
            aspect="auto",
            interpolation="none",
        )

        axs[param_index].set_title(list(evaluator.param_names)[param_index])

        x_ticks_pos = [0, int(max_n_gen / gen_per_bin) - 1]
        x_ticks_label = [0, int(max_n_gen / gen_per_bin) * gen_per_bin]
        axs[param_index].set_xticks(x_ticks_pos, x_ticks_label)
        axs[param_index].set_yticks([0, 19], [param.bounds[1], param.bounds[0]])
        axs[param_index].set_xlim(0, int(max_n_gen / gen_per_bin) - 1)

    for axs_index in range(len(evaluator.params), len(axs)):
        axs[axs_index].set_visible(False)

    # Add a colorbar common to all subplots
    norm = colors.Normalize(vmin=0, vmax=100, clip=False)
    fig.colorbar(
        mappable=cm.ScalarMappable(norm=norm, cmap="viridis"),
        orientation="vertical",
        ax=axs[-1],
        label="% of population",
    )

    fig.supxlabel("Generations", size="xx-large")
    fig.supylabel("Parameter value", size="xx-large")

    suptitle = "Parameter evolution\n"
    if metadata.emodel is not None:
        suptitle += f"e-model = {metadata.emodel}"
    if metadata.iteration is not None:
        suptitle += f" ; iteration = {metadata.iteration}"
    if seed is not None:
        suptitle += f" ; seed = {seed}"
    fig.suptitle(suptitle, size="xx-large")

    figure_name = metadata.as_string(seed=seed)
    figure_name += "__evo_parameter_density.pdf"

    plt.tight_layout()
    plt.subplots_adjust(top=0.92)

    if write_fig:
        save_fig(figures_dir, figure_name)

    return fig, axs


def evolution_parameters_density(
    evaluator, checkpoint_paths, metadata, figures_dir="./figures", write_fig=True
):
    """Create plots of the evolution of the density of parameters in the population as the
    optimisation progresses. Create one plot per checkpoint plus one plot with all checkpoints.

    WARNING: This plotting function assumes that all the checkpoint files come from the same run
    and have the same parameters with the same bounds, therefore, that they come from the same
    evaluator. Do not use otherwise.

    Args:
        evaluator (CellEvaluator): evaluator used to evaluate the individuals.
        checkpoint_paths (list of str): list of paths to the checkpoints .pkl.
        metadata (EModelMetadata): metadata of the emodel.
        figures_dir (str): path to the directory where the figures will be saved.
        write_fig (bool): whether to write the figures to disk.
    """

    make_dir(figures_dir)

    max_n_gen = 0
    genealogies = {}
    for checkpoint_path in checkpoint_paths:
        run, seed = read_checkpoint(checkpoint_path)
        if run["generation"] < 4:
            continue

        max_n_gen = max(max_n_gen, run["generation"])
        genealogies[checkpoint_path] = (run["history"].genealogy_history, seed)

    gen_per_bin = 4
    pop_size = len(run["population"])
    histo_bins = (int(max_n_gen / gen_per_bin), 20)
    normalization_factor = gen_per_bin * pop_size

    # Compute and plot the histograms for each checkpoint
    sum_histograms = {}
    for checkpoint_path, (genealogy, seed) in genealogies.items():
        # Get the histograms for all parameters
        histograms = {}
        for param_index in range(len(genealogy[1])):
            x = [(ind_idx - 1) // pop_size for ind_idx in genealogy.keys()]
            y = [ind[param_index] for ind in genealogy.values()]

            histo_range = [
                [0, max_n_gen],
                [
                    evaluator.params[param_index].bounds[0],
                    evaluator.params[param_index].bounds[1],
                ],
            ]

            h, _, _ = numpy.histogram2d(x, y, bins=histo_bins, range=histo_range)
            normalized_h = h / normalization_factor

            histograms[param_index] = normalized_h
            if param_index not in sum_histograms:
                sum_histograms[param_index] = normalized_h
            else:
                sum_histograms[param_index] = sum_histograms[param_index] + normalized_h

        # Create the figure
        _ = _create_figure_parameter_histograms(
            histograms,
            evaluator,
            metadata,
            seed,
            max_n_gen,
            gen_per_bin,
            figures_dir,
            write_fig,
        )

    # Plot the figure with the sums of all histograms
    fig, axs = None, None
    if sum_histograms:
        sum_histograms = {idx: h / len(checkpoint_path) for idx, h in sum_histograms.items()}
        fig, axs = _create_figure_parameter_histograms(
            sum_histograms,
            evaluator,
            metadata,
            "all_seeds",
            max_n_gen,
            gen_per_bin,
            figures_dir,
            write_fig,
        )

    return fig, axs


def scores(model, figures_dir="./figures", write_fig=True):
    """Plot the scores of a model"""
    SCORE_THRESHOLD = 5.0

    make_dir(figures_dir)

    score = list(model.scores.values()) + list(model.scores_validation.values())
    scores_names = list(model.scores.keys()) + list(model.scores_validation.keys())

    pos = [*range(len(score))]

    fig, axs = plt.subplots(1, figsize=(6, 0.8 + 0.150 * len(pos)), squeeze=False)

    axs[0, 0].barh(pos, score, height=0.7, align="center", color="gray")

    for p, s in zip(pos, score):
        if s > SCORE_THRESHOLD:
            axs[0, 0].text(5.15, p - 0.25, s="{:.1f}".format(s), color="red", fontsize=8)

    axs[0, 0].set_xlabel("z-score")

    axs[0, 0].set_yticks(pos)
    axs[0, 0].set_yticklabels(scores_names, size="small")

    axs[0, 0].set_xlim(0, 5)
    axs[0, 0].set_ylim(-0.5, len(pos) - 0.5)

    title = get_title(model.emodel_metadata.emodel, model.emodel_metadata.iteration, model.seed)
    # tweak size and placement so that title does not overcross figure
    fig.suptitle(title, size="medium", y=0.99)

    fname = model.emodel_metadata.as_string(model.seed) + "__scores.pdf"

    plt.tight_layout()

    if write_fig:
        save_fig(figures_dir, fname)

    return fig, axs


def plot_traces_current(ax, time, current):
    """Plot the current trace on top of the voltage trace"""
    ax.plot(time, current, color="gray", alpha=0.6)

    ax.set_xlabel("Time (ms)")
    ax.set_ylabel("Stim Current (nA)")

    min_lim = numpy.min(current) - 0.2
    max_lim = numpy.max(current) + 0.2
    if numpy.isfinite(min_lim) and numpy.isfinite(max_lim):
        ax.set_ylim(min_lim, max_lim)


def traces_title(model, threshold=None, holding=None, rmp=None, rin=None):
    """Return the title for the traces figure"""
    title = str(model.emodel_metadata.emodel)
    title += f"\n iteration = {model.emodel_metadata.iteration} ; seed = {model.seed}"

    if threshold:
        title += "\n Threshold current = {:.4f} nA".format(threshold)
    if holding:
        title += " ; Holding current = {:.4f} nA".format(holding)
    if rmp:
        title += "\n Resting membrane potential = {:.2f} mV".format(rmp)
    if rin:
        title += " ; Input Resistance = {:.2f} MOhm".format(rin)

    return title


def thumbnail(
    model,
    responses,
    recording_names,
    figures_dir="./figures",
    write_fig=True,
    dpi=300,
    thumbnail_rec=None,
):
    """Plot the trace figure to use as thumbnail."""
    make_dir(figures_dir)

    trace_name = select_rec_for_thumbnail(recording_names, thumbnail_rec=thumbnail_rec)

    # in case e.g. the run fails during preprotocols
    try:
        time = responses[trace_name]["time"]
        voltage = responses[trace_name]["voltage"]
    except KeyError:
        logger.warning(
            "Could not find protocol %s in respsonses. Skipping thumbnail plotting.",
            trace_name,
        )
        return None, None

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)

    ylabel = get_traces_ylabel(var=trace_name.split(".")[-1])
    ax.plot(time, voltage, color="black")
    ax.set_xlabel("Time (ms)")
    ax.set_ylabel(ylabel)

    fname = model.emodel_metadata.as_string(model.seed) + "__thumbnail.png"

    if write_fig:
        save_fig(figures_dir, fname, dpi=dpi)

    return fig, ax


def traces(
    model,
    responses,
    recording_names,
    stimuli={},
    figures_dir="./figures",
    write_fig=True,
):
    """Plot the traces of a model"""
    # pylint: disable=too-many-nested-blocks
    make_dir(figures_dir)

    traces_names, threshold, holding, rmp, rin = get_traces_names_and_float_responses(
        responses, recording_names
    )

    if not traces_names:
        return None, None

    fig, axs = plt.subplots(
        len(traces_names),
        1,
        figsize=(10, 2 + (1.5 * len(traces_names))),
        squeeze=False,
        layout="constrained",
    )

    axs_c = []
    for idx, t in enumerate(sorted(traces_names)):
        axs[idx, 0].set_title(t)

        if responses[t]:
            ylabel = get_traces_ylabel(var=t.split(".")[-1])

            # Plot responses (voltage, current, etc.)
            axs[idx, 0].plot(responses[t]["time"], responses[t]["voltage"], color="black")
            axs[idx, 0].set_xlabel("Time (ms)")
            axs[idx, 0].set_ylabel(ylabel)

            # Plot current
            basename = re.split(r"\.(?=[a-zA-Z])", t, 1)[0]
            if basename in stimuli:
                prot = stimuli[basename]
                if hasattr(prot, "stimulus"):
                    if isinstance(prot, ThresholdBasedProtocol) and threshold and holding:
                        prot.stimulus.holding_current = holding
                        prot.stimulus.threshold_current = threshold
                        if (
                            hasattr(prot, "dependencies")
                            and prot.dependencies["stimulus.holding_current"][1]
                            != "bpo_holding_current"
                        ):
                            resp_name = prot.dependencies["stimulus.holding_current"][1]
                            if resp_name in responses:
                                prot.stimulus.holding_current = responses[resp_name]
                            else:
                                continue
                        if (
                            hasattr(prot, "dependencies")
                            and prot.dependencies["stimulus.threshold_current"][1]
                            != "bpo_threshold_current"
                        ):
                            resp_name = prot.dependencies["stimulus.threshold_current"][1]
                            if resp_name in responses:
                                prot.stimulus.threshold_current = responses[resp_name]
                            else:
                                continue

                    time, current = prot.stimulus.generate()
                    if len(time) > 0 and len(current) > 0:
                        axs_c.append(axs[idx, 0].twinx())
                        plot_traces_current(axs_c[-1], time, current)

        idx += 1

    title = traces_title(model, threshold, holding, rmp, rin)
    fig.suptitle(title)

    fname = model.emodel_metadata.as_string(model.seed) + "__traces.pdf"

    if write_fig:
        save_fig(figures_dir, fname)

    return fig, axs


def dendritic_feature_plot(
    model, responses, feature, feature_name, figures_dir="./figures", write_fig=True
):
    """Plots a accross dendrites and compare it with experimental data.

    Args:
        model (bluepyopt.ephys.CellModel): cell model
        responses (dict): responses of the cell model
        feature (DendFitFeature): feature to plot
        feature_name (str): which feature to plot. Can be either 'ISI_CV' or 'rheobase'.
        figures_dir (str or Path): Where to save the figures.
        write_fig (bool): whether to save the figure

    Returns a figure and its single axe.
    """
    make_dir(figures_dir)

    if feature_name == "ISI_CV":
        exp_label = "exp. data (Shai et al. 2015)"
        y_label = "ISI CV"
        fig_title = "ISI CV along the apical dendrite main branch"
    elif feature_name == "rheobase":
        exp_label = "exp. data (Beaulieu-Laroche (2021))"
        y_label = "rheobase (nA)"
        fig_title = "rheobase along the apical dendrite main branch"
    else:
        raise ValueError(f"Expected 'ISI_CV' or 'rheobase' for feature_name. Got {feature_name}")

    distances, feat_values = feature.get_distances_feature_values(responses)
    exp_distances, exp_values = read_dendritic_data(feature_name)

    min_dist = min((distances[0], exp_distances[0]))
    max_dist = max((distances[-1], exp_distances[-1]))

    # model fit
    if 0 in distances:
        slope = numpy.array([feature.fit(distances, feat_values)])
        x_fit = numpy.linspace(min_dist, max_dist, num=20)
        y_fit = feature.linear_fit(x_fit, slope)

    # data fit
    data_fit = numpy.polyfit(exp_distances, exp_values, 1)
    x_data_fit = numpy.linspace(min_dist, max_dist, num=20)
    y_data_fit = numpy.poly1d(data_fit)(x_data_fit)

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)

    ax.scatter(exp_distances, exp_values, c=colours["datapoint"], label=exp_label)
    ax.plot(x_data_fit, y_data_fit, c=colours["dataline"], label="data fit")
    ax.scatter(distances, feat_values, c=colours["modelpoint_apical"], label="emodel")
    if 0 in distances:
        ax.plot(x_fit, y_fit, "--", c=colours["modelline_apical"], label="emodel fit")
    ax.set_xlabel(r"distance from soma ($\mu$m)")
    ax.set_ylabel(y_label)
    ax.legend(fontsize="x-small")

    fig.suptitle(fig_title)

    if write_fig:
        fname = model.emodel_metadata.as_string(model.seed) + f"__{feature.name}.pdf"
        # can have '[', ']' and ',' in feature name. Replace those characters
        fname = fname.replace("[", "_")
        fname = fname.replace("]", "_")
        fname = fname.replace(",", "_")
        save_fig(figures_dir, fname)

    return fig, ax


def dendritic_feature_plots(mo, feature_name, dest_leaf, figures_dir="./figures"):
    """Calls dendritic_feature_plot for all features corresponding to feature_name.

    Args:
        mo (bluepyopt.ephys.CellModel): cell model
        feature_name (str): which feature to plot. Can be either 'ISI_CV' or 'rheobase'.
        dest_leaf (str): name of repo to use in output path. Usually either 'validated' or 'all'.
        figures_dir (str or Path): base directory where to save the figures.
    """
    figures_dir = Path(figures_dir)
    # translate feature_name into whatever name we expect to be present in fitness_calculator
    if feature_name == "ISI_CV":
        efeature_name = "ISI_CV_linear"
    elif feature_name == "rheobase":
        efeature_name = "bpo_threshold_current_linear"
    else:
        raise ValueError(f"Expected 'ISI_CV' or 'rheobase' for feature_name. Got {feature_name}")

    figures_dir_dendritic = figures_dir / "dendritic" / dest_leaf
    dend_feat_list = [
        obj.features[0]
        for obj in mo.evaluator.fitness_calculator.objectives
        if efeature_name in obj.features[0].name
    ]
    if len(dend_feat_list) < 1:
        logger.debug(
            "Could not find any feature with %s feature name for emodel %s.",
            efeature_name,
            mo.emodel_metadata.emodel,
        )
    for feat in dend_feat_list:
        # prevent stimulus current to be None if load_from_local is True
        # it is not used in this feature computation, but can make the plot crash if None
        if (
            feat.stimulus_current is None
            or callable(feat.stimulus_current)
            and feat.stimulus_current() is None
        ):
            feat.stimulus_current = 0.0

        dendritic_feature_plot(mo, mo.responses, feat, feature_name, figures_dir_dendritic)


def _get_if_curve_from_evaluator(
    holding,
    threshold,
    model,
    evaluator,
    delay,
    length_step,
    delta_current,
    max_offset_current,
):
    total_duration = length_step + (2 * delay)
    stim_end = delay + length_step

    efel.reset()
    efel.set_int_setting("strict_stiminterval", True)

    soma_loc = NrnSeclistCompLocation(name="soma", seclist_name="somatic", sec_index=0, comp_x=0.5)
    rec = CompRecording(name="Step1.soma.v", location=soma_loc, variable="v")

    holding_pulse = NrnSquarePulse(
        step_amplitude=holding,
        step_delay=0.0,
        step_duration=total_duration,
        total_duration=total_duration,
        location=soma_loc,
    )

    spike_freq_equivalent = []
    frequencies = []
    amps = numpy.arange(0.0, threshold + max_offset_current, delta_current)
    for amp in amps:
        step_pulse = NrnSquarePulse(
            step_amplitude=amp,
            step_delay=delay,
            step_duration=length_step,
            total_duration=total_duration,
            location=soma_loc,
        )

        protocol = SweepProtocol("Step1", [holding_pulse, step_pulse], [rec])
        evaluator.fitness_protocols = {"Step1": protocol}

        responses = evaluator.run_protocols(
            protocols=evaluator.fitness_protocols.values(),
            param_values=model.parameters,
        )

        efel_trace = {
            "T": responses["Step1.soma.v"].response["time"],
            "V": responses["Step1.soma.v"].response["voltage"],
            "stim_start": [delay],
            "stim_end": [stim_end],
        }
        features = efel.get_feature_values(
            [efel_trace], ["Spikecount", "mean_frequency"], raise_warnings=False
        )[0]
        spike_freq_equivalent.append(1e3 * float(features["Spikecount"]) / length_step)
        frequencies.append(features.get("mean_frequency", None))

    return amps, frequencies, spike_freq_equivalent


def IF_curve(
    model,
    responses,
    evaluator,
    delay=100,
    length_step=1000,
    delta_current=0.01,
    max_offset_current=0.2,
    figures_dir="./figures",
    write_fig=True,
):
    """Plot the current / frequency curve for the model"""

    fig, ax = plt.subplots(1, figsize=(6, 4))
    ax2 = ax.twinx()

    holding = responses.get("bpo_holding_current", None)
    threshold = responses.get("bpo_threshold_current", None)
    if holding is None or threshold is None:
        logger.warning("Not plotting IF curve, holding or threshold current is missing")
        return fig, [ax, ax2]

    amps, frequencies, spike_freq_equivalent = _get_if_curve_from_evaluator(
        holding,
        threshold,
        model,
        evaluator,
        delay,
        length_step,
        delta_current,
        max_offset_current,
    )

    ax.scatter(amps, frequencies, c="C0", alpha=0.6)
    ax.set_xlabel("Step amplitude (nA)")
    ax.set_ylabel("Mean frequency (Hz)", color="C0")
    ax.tick_params(axis="y", labelcolor="C0")

    ax2.scatter(amps, spike_freq_equivalent, c="C1", alpha=0.6)
    ax2.set_ylabel("Spikecount per s over the step", color="C1")
    ax2.tick_params(axis="y", labelcolor="C1")

    title = f"IF curve {model.emodel_metadata.emodel}, seed {model.seed}"

    fig.suptitle(title)

    fname = model.emodel_metadata.as_string(model.seed) + "__IF_curve.pdf"

    plt.tight_layout()

    if write_fig:
        save_fig(figures_dir, fname)

    return fig, [ax, ax2]


def parameters_distribution(models, lbounds, ubounds, figures_dir="./figures", write_fig=True):
    """Plot the distribution of the parameters across several models"""
    make_dir(figures_dir)

    if len({mo.emodel_metadata.as_string() for mo in models}) > 1:
        logger.warning(
            "More than one e-type passed to the plotting.parameters_distribution function"
        )

    # Normalizes the parameters and makes sure they are listed in the same order
    data = []
    parameters = list(lbounds.keys())
    for mo in models:
        _ = []
        for param in parameters:
            bm = (ubounds[param] + lbounds[param]) / 2.0
            br = (ubounds[param] - lbounds[param]) / 2.0
            _.append((mo.parameters[param] - bm) / br)
        data.append(_)
    data = numpy.array(data)

    figure_width = 2.5 + 0.08 * max(len(pn) for pn in ubounds.keys())
    figure_height = 0.8 + 0.21 * len(ubounds)
    fig, axs = plt.subplots(1, figsize=(figure_width, figure_height), squeeze=False)

    v = axs[0, 0].violinplot(data, vert=False)

    for partname in ("cbars", "cmins", "cmaxes", "cmeans", "cmedians"):
        if partname in v:
            vp = v[partname]
            vp.set_edgecolor("black")
    for pc in v["bodies"]:
        pc.set_facecolor("black")

    axs[0, 0].set_yticks(ticks=range(1, 1 + len(ubounds)))
    axs[0, 0].set_yticklabels(labels=list(ubounds.keys()))

    # Guides for the eye
    for p in range(len(ubounds)):
        axs[0, 0].plot([-1, 1], [p + 1, p + 1], c="black", ls="dotted", alpha=0.7, lw=0.7, zorder=1)
    axs[0, 0].plot([-1, -1], [0, 1 + len(ubounds)], c="black", ls="--", alpha=0.8, zorder=1)
    axs[0, 0].plot([0, 0], [0, 1 + len(ubounds)], c="black", ls="--", alpha=0.8, zorder=1)
    axs[0, 0].plot([1, 1], [0, 1 + len(ubounds)], c="black", ls="--", alpha=0.8, zorder=1)

    axs[0, 0].set_xticks(ticks=[-1, 0, 1])
    axs[0, 0].set_xticklabels(labels=["Lower bounds", "50%", "Upper bounds"], rotation=90)

    axs[0, 0].set_ylim(0, 1 + len(ubounds))
    axs[0, 0].set_xlim(-1.05, 1.05)

    for _, spine in axs[0, 0].spines.items():
        spine.set_visible(False)

    title = str(models[0].emodel_metadata.emodel)
    title += f"; iteration = {models[0].emodel_metadata.iteration}"
    axs[0, 0].set_title(title)

    fname = models[0].emodel_metadata.as_string() + "__parameters_distribution.pdf"

    plt.tight_layout()

    if write_fig:
        save_fig(figures_dir, fname)

    return fig, axs


def bAP_fit(feature, distances, values, npoints=20):
    """Returns a x and y arrays, y being a exponential decay fit for bAP feature.

    Args:
        feature (DendFitFeature): bAP feature
        distances (list): distances of the recordings
        values (list): bAP values
        npoints (int): number of items in the returned ndarrays
    """
    slope = numpy.array([feature.fit(distances, values)])
    x_fit = numpy.linspace(distances[0], distances[-1], num=npoints)
    y_fit = feature.exp_decay(x_fit, slope)
    return x_fit, y_fit


def EPSP_fit(feature, distances, values, npoints=20):
    """Returns a x and y arrays, y being a exponential fit for EPSP feature.

    Args:
        feature (DendFitFeature): EPSP feature
        distances (list): distances of the recordings
        values (list): EPSP values
        npoints (int): number of items in the returned ndarrays
    """
    slope = numpy.array([feature.fit(distances, values)])
    x_fit = numpy.linspace(distances[0], distances[-1], num=npoints)
    y_fit = feature.exp(x_fit, slope)
    return x_fit, y_fit


def plot_bAP(
    model,
    responses,
    apical_feature,
    basal_feature,
    figures_dir="./figures",
    write_fig=True,
):
    """Plot back-propagating action potential.

    Args:
        model (bluepyopt.ephys.CellModel): cell model
        responses (dict): responses of the cell model
        apical_feature (DendFitFeature): bAP feature with apical recs,
        basal_feature (DendFitFeature): bAP feature with basal recs,
        figures_dir (str or Path): directory where to save the figures
        write_fig (bool): whether to save the figure
    """
    make_dir(figures_dir)

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)

    apical_distances, apical_values = apical_feature.get_distances_feature_values(responses)
    basal_distances, basal_values = basal_feature.get_distances_feature_values(responses)

    if 0 in apical_distances:
        apical_x_fit, apical_y_fit = bAP_fit(apical_feature, apical_distances, apical_values)
    if 0 in basal_distances:
        basal_x_fit, basal_y_fit = bAP_fit(basal_feature, basal_distances, basal_values)

    ax.scatter(
        apical_distances,
        apical_values,
        c=colours["modelpoint_apical"],
        label="model apical",
    )
    ax.scatter(
        basal_distances,
        basal_values,
        c=colours["modelpoint_basal"],
        label="model basal",
    )
    if 0 in apical_distances:
        ax.plot(
            apical_x_fit,
            apical_y_fit,
            "--",
            c=colours["modelline_apical"],
            label="model apical fit",
        )
    if 0 in basal_distances:
        ax.plot(
            basal_x_fit,
            basal_y_fit,
            "--",
            c=colours["modelline_basal"],
            label="model basal fit",
        )
    ax.set_xlabel(r"Distance from soma ($\mu$m)")
    ax.set_ylabel("Amplitude (mV)")
    ax.legend(fontsize="x-small")

    fig.suptitle("Dendrite back-propagating action potential")

    if write_fig:
        fname = (
            model.emodel_metadata.as_string(model.seed) + "__dendrite_backpropagation_fit_decay.pdf"
        )
        save_fig(figures_dir, fname)

    return fig, ax


def compute_attenuation(dendrec_feature, somarec_feature, responses):
    """Returns EPSP attenuation and corresponding distances.

    Args:
        dendrec_feature (DendFitFeature): feature with recordings in the dendrite
        somarec_feature (DendFitFeature): feature with recordings in the soma
        responses (dict): responses to feed to the features
    """
    distances, dend_values = dendrec_feature.get_distances_feature_values(responses)
    _, soma_values = somarec_feature.get_distances_feature_values(responses)
    attenuation = numpy.asarray(dend_values) / numpy.asarray(soma_values)

    return distances, attenuation


def plot_EPSP(
    model,
    responses,
    apical_apicrec_feat,
    apical_somarec_feat,
    basal_basalrec_feat,
    basal_somarec_feat,
    figures_dir="./figures",
    write_fig=True,
):
    """Plot EPSP attenuation across dendrites.

    Args:
        model (bluepyopt.ephys.CellModel): cell model
        responses (dict): responses of the cell model
        apical_apicrec_feat (DendFitFeature): EPSP feature with apical stim and apical recs,
        apical_somarec_feat (DendFitFeature): EPSP feature with apical stim and soma recs,
        basal_basalrec_feat (DendFitFeature): EPSP feature with basal stim and basal recs,
        basal_somarec_feat (DendFitFeature): EPSP feature with basal stim and soma recs,
        figures_dir (str or Path): directory where to save the figures
        write_fig (bool): whether to save the figure
    """
    make_dir(figures_dir)

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)

    apical_distances, apical_attenuation = compute_attenuation(
        apical_apicrec_feat, apical_somarec_feat, responses
    )
    basal_distances, basal_attenuation = compute_attenuation(
        basal_basalrec_feat, basal_somarec_feat, responses
    )

    if 0 in apical_distances:
        apical_x_fit, apical_y_fit = EPSP_fit(
            apical_apicrec_feat, apical_distances, apical_attenuation
        )
    if 0 in basal_distances:
        basal_x_fit, basal_y_fit = EPSP_fit(basal_basalrec_feat, basal_distances, basal_attenuation)

    ax.scatter(
        apical_distances,
        apical_attenuation,
        c=colours["modelpoint_apical"],
        label="model apical",
    )
    ax.scatter(
        basal_distances,
        basal_attenuation,
        c=colours["modelpoint_basal"],
        label="model basal",
    )
    if 0 in apical_distances:
        ax.plot(
            apical_x_fit,
            apical_y_fit,
            "--",
            c=colours["modelline_apical"],
            label="model apical fit",
        )
    if 0 in basal_distances:
        ax.plot(
            basal_x_fit,
            basal_y_fit,
            "--",
            c=colours["modelline_basal"],
            label="model basal fit",
        )
    ax.set_xlabel(r"Distance from soma ($\mu$m)")
    ax.set_ylabel("Attenuation dendrite amplitude / soma amplitude")
    ax.legend(fontsize="x-small")

    fig.suptitle("EPSP attenuation")

    if write_fig:
        fname = model.emodel_metadata.as_string(model.seed) + "__dendrite_EPSP_attenuation_fit.pdf"
        save_fig(figures_dir, fname)

    return fig, ax


def run_and_plot_bAP(
    original_cell_evaluator,
    access_point,
    mapper,
    seeds,
    save_recordings,
    load_from_local,
    only_validated=False,
    figures_dir="./figures",
):
    """Runs and plots ready-to-use bAP protocol for apical and basal dendrites.

    Args:
        original_cell_evaluator (CellEvaluator): original cell evaluator.
            A copy will be modified and then used to compute the responses.
        access_point (DataAccessPoint): data access point.
        mapper (map): used to parallelize the evaluation of the
            individual in the population.
        seeds (list): if not None, filter emodels to keep only the ones with these seeds.
        save_recordings (bool): Whether to save the responses data under a folder
            named `recordings`. Responses can then be loaded using load_from_local
            instead of being re-run.
        load_from_local (bool): True to load responses from locally saved recordings.
            Responses are saved locally when save_recordings is True.
        only_validated (bool): True to only plot validated models
        figures_dir (str): path of the directory in which the figures should be saved.
    """
    figures_dir = Path(figures_dir)
    cell_evaluator = copy.deepcopy(original_cell_evaluator)

    # get basal and apical lengths
    morph_path = cell_evaluator.cell_model.morphology.morphology_path
    max_basal_length, max_apical_length = get_basal_and_apical_max_radial_distances(morph_path)
    max_basal_length = int(max_basal_length)
    max_apical_length = int(max_apical_length)

    # remove protocols except for pre-protocols
    old_prots = cell_evaluator.fitness_protocols["main_protocol"].protocols
    new_prots = {}
    for k, v in old_prots.items():
        if k in PRE_PROTOCOLS:
            new_prots[k] = v

    bAP_prot = define_bAP_protocol(dist_end=max_apical_length, dist_end_basal=max_basal_length)
    new_prots[bAP_prot.name] = bAP_prot

    cell_evaluator.fitness_protocols["main_protocol"].protocols = new_prots
    cell_evaluator.fitness_protocols["main_protocol"].execution_order = (
        cell_evaluator.fitness_protocols["main_protocol"].compute_execution_order()
    )

    apical_feature = define_bAP_feature("apical", dist_end=max_apical_length)
    basal_feature = define_bAP_feature("basal", dist_end=max_basal_length)
    # run emodel(s)
    emodels = compute_responses(
        access_point,
        cell_evaluator,
        mapper,
        seeds,
        store_responses=save_recordings,
        load_from_local=load_from_local,
    )

    if only_validated:
        emodels = [model for model in emodels if model.passed_validation]
        dest_leaf = "validated"
    else:
        dest_leaf = "all"

    # plot
    for mo in emodels:
        figures_dir_bAP = figures_dir / "dendritic" / dest_leaf
        plot_bAP(mo, mo.responses, apical_feature, basal_feature, figures_dir_bAP)


def run_and_plot_EPSP(
    original_cell_evaluator,
    access_point,
    mapper,
    seeds,
    save_recordings,
    load_from_local,
    only_validated=False,
    figures_dir="./figures",
):
    """Runs and plots ready-to-use EPSP protocol for apical and basal dendrites.

    Args:
        original_cell_evaluator (CellEvaluator): original cell evaluator.
            A copy will be modified and then used to compute the responses.
        access_point (DataAccessPoint): data access point.
        mapper (map): used to parallelize the evaluation of the
            individual in the population.
        seeds (list): if not None, filter emodels to keep only the ones with these seeds.
        save_recordings (bool): Whether to save the responses data under a folder
            named `recordings`. Responses can then be loaded using load_from_local
            instead of being re-run.
        load_from_local (bool): True to load responses from locally saved recordings.
            Responses are saved locally when save_recordings is True.
        only_validated (bool): True to only plot validated models
        figures_dir (str): path of the directory in which the figures should be saved.
    """
    figures_dir = Path(figures_dir)
    cell_evaluator = copy.deepcopy(original_cell_evaluator)

    # get basal and apical lengths
    morph_path = cell_evaluator.cell_model.morphology.morphology_path
    max_basal_length, max_apical_length = get_basal_and_apical_max_radial_distances(morph_path)
    max_basal_length = int(max_basal_length)
    max_apical_length = int(max_apical_length)

    # remove protocols except for pre-protocols
    apical_prots = define_EPSP_protocol(
        "apical", dist_end=max_apical_length, dist_start=100, dist_step=100
    )
    basal_prots = define_EPSP_protocol(
        "basal", dist_end=max_basal_length, dist_start=30, dist_step=30
    )
    EPSP_prots = {**apical_prots, **basal_prots}

    cell_evaluator.fitness_protocols["main_protocol"].protocols = EPSP_prots
    cell_evaluator.fitness_protocols["main_protocol"].execution_order = (
        cell_evaluator.fitness_protocols["main_protocol"].compute_execution_order()
    )

    apical_apicrec_feat = define_EPSP_feature(
        "apical", "dend", dist_end=max_apical_length, dist_start=100, dist_step=100
    )
    apical_somarec_feat = define_EPSP_feature(
        "apical", "soma", dist_end=max_apical_length, dist_start=100, dist_step=100
    )
    basal_basalrec_feat = define_EPSP_feature(
        "basal", "dend", dist_end=max_basal_length, dist_start=30, dist_step=30
    )
    basal_somarec_feat = define_EPSP_feature(
        "basal", "soma", dist_end=max_basal_length, dist_start=30, dist_step=30
    )
    # run emodel(s)
    emodels = compute_responses(
        access_point,
        cell_evaluator,
        mapper,
        seeds,
        store_responses=save_recordings,
        load_from_local=load_from_local,
    )

    if only_validated:
        emodels = [model for model in emodels if model.passed_validation]
        dest_leaf = "validated"
    else:
        dest_leaf = "all"

    # plot
    for mo in emodels:
        figures_dir_EPSP = figures_dir / "dendritic" / dest_leaf
        plot_EPSP(
            mo,
            mo.responses,
            apical_apicrec_feat,
            apical_somarec_feat,
            basal_basalrec_feat,
            basal_somarec_feat,
            figures_dir_EPSP,
        )


def plot_models(
    access_point,
    mapper,
    seeds=None,
    figures_dir="./figures",
    plot_distributions=True,
    plot_scores=True,
    plot_traces=True,
    plot_thumbnail=True,
    plot_currentscape=False,
    plot_if_curve=False,
    plot_dendritic_ISI_CV=True,
    plot_dendritic_rheobase=True,
    plot_bAP_EPSP=False,
    only_validated=False,
    save_recordings=False,
    load_from_local=False,
    cell_evaluator=None,
):
    """Plot the traces, scores and parameter distributions for all the models
        matching the emodels name.

    Args:
        access_point (DataAccessPoint): data access point.
        mapper (map): used to parallelize the evaluation of the
            individual in the population.
        seeds (list): if not None, filter emodels to keep only the ones with these seeds.
        figures_dir (str): path of the directory in which the figures should be saved.
        plot_distributions (bool): True to plot the parameters distributions
        plot_scores (bool): True to plot the scores
        plot_traces (bool): True to plot the traces
        plot_thumbnail (bool): True to plot a trace used as thumbnail
        plot_currentscape (bool): True to plot the currentscapes
        plot_if_curve (bool): True to plot the current / frequency curve
        plot_dendritic_ISI_CV (bool): True to plot dendritic ISI CV (if present)
        plot_dendritic_rheobase (bool): True to plot dendritic rheobase (if present)
        plot_bAP_EPSP (bool): True to plot bAP and EPSP protocol.
            Only use this on model having apical dendrites.
        only_validated (bool): True to only plot validated models
        save_recordings (bool): Whether to save the responses data under a folder
            named `recordings`. Responses can then be loaded using load_from_local
            instead of being re-run.
        load_from_local (bool): True to load responses from locally saved recordings.
            Responses are saved locally when save_recordings is True.
        cell_evaluator (CellEvaluator): cell evaluator used to compute the responses.

    Returns:
        emodels (list): list of emodels.
    """
    # pylint: disable=too-many-arguments, too-many-locals

    figures_dir = Path(figures_dir)

    if cell_evaluator is None:
        cell_evaluator = get_evaluator_from_access_point(
            access_point,
            include_validation_protocols=True,
            record_ions_and_currents=plot_currentscape,
        )

    if (
        plot_currentscape
        and access_point.pipeline_settings.currentscape_config is not None
        and "current" in access_point.pipeline_settings.currentscape_config
        and "names" in access_point.pipeline_settings.currentscape_config["current"]
    ):
        add_recordings_to_evaluator(
            cell_evaluator,
            access_point.pipeline_settings.currentscape_config["current"]["names"],
            use_fixed_dt_recordings=False,
        )

    if (
        plot_traces
        or plot_currentscape
        or plot_dendritic_ISI_CV
        or plot_dendritic_rheobase
        or plot_thumbnail
    ):
        emodels = compute_responses(
            access_point,
            cell_evaluator,
            mapper,
            seeds,
            store_responses=save_recordings,
            load_from_local=load_from_local,
        )
    else:
        emodels = access_point.get_emodels([access_point.emodel_metadata.emodel])
        if seeds:
            emodels = [model for model in emodels if model.seed in seeds]

    stimuli = cell_evaluator.fitness_protocols["main_protocol"].protocols

    if only_validated:
        emodels = [model for model in emodels if model.passed_validation]
        dest_leaf = "validated"
    else:
        dest_leaf = "all"

    if not emodels:
        logger.warning("In plot_models, no emodel for %s", access_point.emodel_metadata.emodel)
        return []

    if plot_distributions:
        lbounds = {
            p.name: p.bounds[0]
            for p in cell_evaluator.cell_model.params.values()
            if p.bounds is not None
        }
        ubounds = {
            p.name: p.bounds[1]
            for p in cell_evaluator.cell_model.params.values()
            if p.bounds is not None
        }

        figures_dir_dist = figures_dir / "distributions" / dest_leaf

        parameters_distribution(
            models=emodels,
            lbounds=lbounds,
            ubounds=ubounds,
            figures_dir=figures_dir_dist,
        )

    for mo in emodels:
        if plot_scores:
            figures_dir_scores = figures_dir / "scores" / dest_leaf
            scores(mo, figures_dir_scores)
        if plot_traces:
            figures_dir_traces = figures_dir / "traces" / dest_leaf
            recording_names = get_recording_names(
                access_point.get_fitness_calculator_configuration().protocols,
                stimuli,
            )
            traces(mo, mo.responses, recording_names, stimuli, figures_dir_traces)

        if plot_thumbnail:
            figures_dir_thumbnail = figures_dir / "thumbnail" / dest_leaf
            recording_names = get_recording_names(
                access_point.get_fitness_calculator_configuration().protocols,
                stimuli,
            )
            thumbnail(mo, mo.responses, recording_names, figures_dir_thumbnail)

        if plot_dendritic_ISI_CV:
            dendritic_feature_plots(mo, "ISI_CV", dest_leaf, figures_dir)

        if plot_dendritic_rheobase:
            dendritic_feature_plots(mo, "rheobase", dest_leaf, figures_dir)

        if plot_currentscape:
            config = access_point.pipeline_settings.currentscape_config
            figures_dir_currentscape = figures_dir / "currentscape" / dest_leaf
            currentscape(
                mo.responses,
                config=config,
                metadata_str=mo.emodel_metadata.as_string(mo.seed),
                figures_dir=figures_dir_currentscape,
                emodel=mo.emodel_metadata.emodel,
            )
            # reset rcParams which are modified by currentscape
            matplotlib.rcParams.update(matplotlib.rcParamsDefault)

        if plot_if_curve:
            figures_dir_traces = figures_dir / "traces" / dest_leaf
            IF_curve(
                mo,
                mo.responses,
                copy.deepcopy(cell_evaluator),
                figures_dir=figures_dir_traces,
            )

    if plot_bAP_EPSP:
        run_and_plot_bAP(
            cell_evaluator,
            access_point,
            mapper,
            seeds,
            save_recordings,
            load_from_local,
            only_validated,
            figures_dir,
        )
        run_and_plot_EPSP(
            cell_evaluator,
            access_point,
            mapper,
            seeds,
            save_recordings,
            load_from_local,
            only_validated,
            figures_dir,
        )

    return emodels


def get_ordered_currentscape_keys(keys):
    """Get responses keys (also filename strings) ordered by protocols and locations.

    Arguments:
        keys (list of str): list of responses keys (or filename stems).
            Each item should have the shape protocol.location.current

    Returns:
        dict: containing voltage key, ion current and ionic concentration keys,
                and ion current and ionic concentration names. Should have the shape:

            .. code-block::

                {
                    "protocol_name": {
                        "loc_name": {
                            "voltage_key": str, "current_keys": [], "current_names": [],
                            "ion_conc_keys": [], "ion_conc_names": [],
                        }
                    }
                }
    """
    # RMP and Rin only have voltage data, no currents, so they are skipped
    to_skip = [
        "RMPProtocol",
        "RinProtocol",
        "SearchHoldingCurrent",
        "bpo_rmp",
        "bpo_rin",
        "bpo_holding_current",
        "bpo_threshold_current",
    ]

    ordered_keys = {}
    for name in keys:
        n = name.split(".")
        # case where protocol has '.' in its name, e.g. IV_-100.0
        if len(n) == 4 and n[1].isdigit():
            n = [".".join(n[:2]), n[2], n[3]]
        prot_name = n[0]
        # prot_name can be e.g. RMPProtocol, or RMPProtocol_apical055
        if not any(to_skip_ in prot_name for to_skip_ in to_skip):
            if len(n) != 3:
                raise ValueError(f"Expected 3 elements in {n}")
            loc_name = n[1]
            curr_name = n[2]

            if prot_name not in ordered_keys:
                ordered_keys[prot_name] = {}
            if loc_name not in ordered_keys[prot_name]:
                ordered_keys[prot_name][loc_name] = {
                    "voltage_key": None,
                    "current_keys": [],
                    "current_names": [],
                    "ion_conc_keys": [],
                    "ion_conc_names": [],
                }

            if curr_name == "v":
                ordered_keys[prot_name][loc_name]["voltage_key"] = name
            elif curr_name[0] == "i":
                ordered_keys[prot_name][loc_name]["current_keys"].append(name)
                ordered_keys[prot_name][loc_name]["current_names"].append(curr_name)
            elif curr_name[-1] == "i":
                ordered_keys[prot_name][loc_name]["ion_conc_keys"].append(name)
                ordered_keys[prot_name][loc_name]["ion_conc_names"].append(curr_name)

    return ordered_keys


def get_voltage_currents_from_files(key_dict, output_dir):
    """Get time, voltage, currents and ionic concentrations from output files"""
    v_path = Path(output_dir) / ".".join((key_dict["voltage_key"], "dat"))
    time = numpy.loadtxt(v_path)[:, 0]
    voltage = numpy.loadtxt(v_path)[:, 1]

    curr_paths = [
        Path(output_dir) / ".".join((curr_key, "dat")) for curr_key in key_dict["current_keys"]
    ]
    currents = [numpy.loadtxt(curr_path)[:, 1] for curr_path in curr_paths]

    ion_conc_paths = [
        Path(output_dir) / ".".join((ion_conc_key, "dat"))
        for ion_conc_key in key_dict["ion_conc_keys"]
    ]
    ionic_concentrations = [numpy.loadtxt(ion_conc_path)[:, 1] for ion_conc_path in ion_conc_paths]

    return time, voltage, currents, ionic_concentrations


def currentscape(
    responses=None,
    output_dir=None,
    config=None,
    metadata_str="",
    figures_dir="./figures",
    emodel="",
):
    """Plot the currentscapes for all protocols.

    Arguments:
        responses (dict): dict containing the current and voltage responses.
        output_dur (str): path to the output dir containing the voltage and current responses.
            Will not be used if responses is set.
        config (dict): currentscape config. See currentscape package for more info.
        metadata_str (str): Metadata of the model as a string. Used in the files naming.
        figures_dir (str): path to the directory where to put the figures.
        emodel (str): name of the emodel
        iteration_tag (str): githash
        seed (int): random seed number
    """
    # pylint: disable=too-many-branches, too-many-statements
    # TODO: refactor this a little
    if responses is None and output_dir is None:
        raise TypeError("Responses or output directory must be set.")

    make_dir(figures_dir)

    if config is None:
        config = {}
    # copy dict so that changes to dict won't affect next call to this function
    updated_config = copy.deepcopy(config)

    if "current" not in updated_config:
        updated_config["current"] = {}
    if "ions" not in updated_config:
        updated_config["ions"] = {}
    if "output" not in updated_config:
        updated_config["output"] = {}
    if "legend" not in updated_config:
        updated_config["legend"] = {}
    if "show" not in updated_config:
        updated_config["show"] = {}

    current_subset = None
    if "names" in updated_config["current"]:
        current_subset = updated_config["current"]["names"].copy()

    if responses is not None:
        ordered_keys = get_ordered_currentscape_keys(
            key for key, item in responses.items() if item is not None
        )
    else:
        fnames = [
            str(Path(filepath).stem) for filepath in glob.glob(str(Path(output_dir) / "*.dat"))
        ]
        ordered_keys = get_ordered_currentscape_keys(fnames)

    for prot, locs in ordered_keys.items():
        for loc, key_dict in locs.items():
            if key_dict["voltage_key"] is None:
                continue
            if responses is not None:
                time = responses[key_dict["voltage_key"]]["time"]
                voltage = responses[key_dict["voltage_key"]]["voltage"]
                # current data has also voltage for a key
                currents = [responses[key]["voltage"] for key in key_dict["current_keys"]]
                ionic_concentrations = [
                    responses[key]["voltage"] for key in key_dict["ion_conc_keys"]
                ]
            else:
                time, voltage, currents, ionic_concentrations = get_voltage_currents_from_files(
                    key_dict, output_dir
                )

            name = ".".join((f"{metadata_str}__currentscape", prot, loc))

            # adapt config
            if current_subset and key_dict["current_names"]:
                try:
                    currents_indices = [
                        list(key_dict["current_names"]).index(c_name) for c_name in current_subset
                    ]
                    currents = numpy.array(currents)[currents_indices]
                    updated_config["current"]["names"] = current_subset
                except ValueError:
                    logger.warning(
                        "Recorded currents do not match current names given in config. "
                        "Skipping currentscape plotting."
                    )
                    # skip plotting by having an empty current list
                    currents = numpy.array([])
                    updated_config["current"]["names"] = []
            else:
                updated_config["current"]["names"] = key_dict["current_names"]
            updated_config["ions"]["names"] = key_dict["ion_conc_names"]
            updated_config["output"]["savefig"] = True
            updated_config["output"]["fname"] = name
            if "dir" not in updated_config["output"]:
                updated_config["output"]["dir"] = figures_dir
            if "title" not in config and emodel:
                # check config because we want to change this for each plot
                title = f"{emodel}\n{prot}"
                updated_config["title"] = title
            # resizing
            if "figsize" not in updated_config:
                updated_config["figsize"] = (4.5, 6)
            if "textsize" not in updated_config:
                updated_config["textsize"] = 8
            if "textsize" not in updated_config["legend"]:
                updated_config["legend"]["textsize"] = 5

            if len(voltage) == 0 or len(currents) == 0:
                logger.warning(
                    "Could not plot currentscape for %s: voltage or currents is empty.",
                    name,
                )
            else:
                try:
                    from currentscape.currentscape import plot_currentscape as plot_currentscape_fct

                    logger.info("Plotting currentscape for %s", name)
                    fig = plot_currentscape_fct(
                        voltage,
                        currents,
                        updated_config,
                        ions_data=ionic_concentrations,
                        time=time,
                    )
                    plt.close(fig)
                except ModuleNotFoundError:
                    logger.warning("Currentscape module not found. Skipping currentscape plotting.")



"""EModel class"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy

from bluepyemodel.tools import search_pdfs


def format_dict_for_resource(d):
    """Translates a dictionary to a list of the format used by resources"""

    out = []

    if d is None:
        return out

    for k, v in d.items():
        if v is None or numpy.isnan(v):
            v = None

        out.append({"name": k, "value": v, "unitCode": ""})

    return out


class EModelMixin:
    """Contains functions used both in EModel and MEModel classes."""

    def build_pdf_dependencies(self, seed):
        """Find all the pdfs associated to an emodel"""

        pdfs = []

        opt_pdf = search_pdfs.search_figure_emodel_optimisation(self.emodel_metadata, seed)
        if opt_pdf:
            pdfs.append(opt_pdf)

        traces_pdf = search_pdfs.search_figure_emodel_traces(self.emodel_metadata, seed)
        if traces_pdf:
            pdfs += [p for p in traces_pdf if p]

        scores_pdf = search_pdfs.search_figure_emodel_score(self.emodel_metadata, seed)
        if scores_pdf:
            pdfs += [p for p in scores_pdf if p]

        thumbnail_pdf = search_pdfs.search_figure_emodel_thumbnail(self.emodel_metadata, seed)
        if thumbnail_pdf:
            pdfs += [p for p in thumbnail_pdf if p]

        parameters_pdf = search_pdfs.search_figure_emodel_parameters(self.emodel_metadata)
        if parameters_pdf:
            pdfs += [p for p in parameters_pdf if p]

        parameters_evo_pdf = search_pdfs.search_figure_emodel_parameters_evolution(
            self.emodel_metadata, seed
        )
        if parameters_evo_pdf:
            pdfs.append(parameters_evo_pdf)

        all_parameters_evo_pdf = search_pdfs.search_figure_emodel_parameters_evolution(
            self.emodel_metadata, seed=None
        )
        if all_parameters_evo_pdf:
            pdfs.append(all_parameters_evo_pdf)

        currentscape_pdfs = search_pdfs.search_figure_emodel_currentscapes(
            self.emodel_metadata, seed
        )
        if currentscape_pdfs:
            pdfs += [p for p in currentscape_pdfs if p]

        bAP_pdf = search_pdfs.search_figure_emodel_bAP(self.emodel_metadata, seed)
        if bAP_pdf:
            pdfs += [p for p in bAP_pdf if p]

        EPSP_pdf = search_pdfs.search_figure_emodel_EPSP(self.emodel_metadata, seed)
        if EPSP_pdf:
            pdfs += [p for p in EPSP_pdf if p]

        ISI_CV_pdf = search_pdfs.search_figure_emodel_ISI_CV(self.emodel_metadata, seed)
        if ISI_CV_pdf:
            pdfs += [p for p in ISI_CV_pdf if p]

        rheobase_pdf = search_pdfs.search_figure_emodel_rheobase(self.emodel_metadata, seed)
        if rheobase_pdf:
            pdfs += [p for p in rheobase_pdf if p]

        return pdfs


class EModel(EModelMixin):
    """Contains all the information related to an optimized e-model, such as its parameters or
    its e-feature values and scores.

    This class is not meant to be instantiated by hand. It is used to transmit the information
    about optimized e-model information."""

    def __init__(
        self,
        fitness=None,
        parameter=None,
        score=None,
        features=None,
        scoreValidation=None,
        passedValidation=None,
        seed=None,
        emodel_metadata=None,
        workflow_id=None,
    ):
        """Init

        Args:
            fitness (float): fitness of the emodel as the sum of scores.
            parameter (dict or Resource): parameters of the emodel
            score (dict or Resource): scores of the emodel.
            features (dict or Resource): feature values of the emodel.
            scoreValidation (dict or Resource): scores obtained on the validation protocols.
            passedValidation (bool or None): did the model go through validation and if yes,
                did it pass it successfully (None: no validation, True: passed, False: didn't pass)
            seed (int): seed used during optimisation for this emodel.
            emodel_metadata (EModelMetadata): metadata of the model (emodel name, etype, ttype, ...)
            workflow_id (str): EModelWorkflow id on nexus.
        """

        self.emodel_metadata = emodel_metadata
        self.passed_validation = passedValidation
        self.fitness = fitness
        self.seed = seed

        self.workflow_id = workflow_id

        if isinstance(parameter, dict):
            self.parameters = parameter
        else:
            self.parameters = {p["name"]: p["value"] for p in parameter} if parameter else {}

        if isinstance(score, dict):
            self.scores = score
        else:
            self.scores = {p["name"]: p["value"] for p in score} if score else {}

        if isinstance(features, dict):
            self.features = features
        else:
            self.features = (
                {p["name"]: (p["value"] if "value" in p else numpy.nan) for p in features}
                if features
                else {}
            )

        if isinstance(scoreValidation, dict):
            self.scores_validation = scoreValidation
        else:
            self.scores_validation = (
                {p["name"]: p["value"] for p in scoreValidation} if scoreValidation else {}
            )

        self.responses = {}
        self.evaluator = None

    def copy_pdf_dependencies_to_new_path(self, seed, overwrite=False):
        """Copy pdf dependencies to new path using allen notation"""
        search_pdfs.copy_emodel_pdf_dependencies_to_new_path(
            self.emodel_metadata, self.emodel_metadata, False, True, seed, overwrite=overwrite
        )

    def get_related_nexus_ids(self):
        return {
            "generation": {
                "type": "Generation",
                "activity": {
                    "type": "Activity",
                    "followedWorkflow": {
                        "type": "EModelWorkflow",
                        "id": self.workflow_id,
                    },
                },
            }
        }

    def as_dict(self):
        scores_validation_resource = format_dict_for_resource(self.scores_validation)
        scores_resource = format_dict_for_resource(self.scores)
        features_resource = format_dict_for_resource(self.features)
        parameters_resource = format_dict_for_resource(self.parameters)
        pdf_dependencies = self.build_pdf_dependencies(self.seed)

        return {
            "fitness": sum(list(self.scores.values())),
            "parameter": parameters_resource,
            "score": scores_resource,
            "features": features_resource,
            "scoreValidation": scores_validation_resource,
            "passedValidation": self.passed_validation,
            "nexus_images": pdf_dependencies,
            "seed": self.seed,
        }



"""TargetsConfiguration"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import itertools
import logging

from bluepyefe.auto_targets import AutoTarget

from bluepyemodel.efeatures_extraction.target import Target
from bluepyemodel.efeatures_extraction.trace_file import TraceFile
from bluepyemodel.tools.utils import are_same_protocol
from bluepyemodel.tools.utils import format_protocol_name_to_list

logger = logging.getLogger(__name__)


class TargetsConfiguration:
    """The goal of this class is to configure the targets and files metadata that will be
    used during efeature extraction"""

    def __init__(
        self,
        files=None,
        targets=None,
        protocols_rheobase=None,
        available_traces=None,
        available_efeatures=None,
        auto_targets=None,
        additional_fitness_efeatures=None,
        additional_fitness_protocols=None,
        protocols_mapping=None,
    ):
        """Init

        Args:
            files (list of dict): File names with their metadata in the format:

                .. code-block::

                    [file1_metadata, file2_metadata, ...]
                    file1_metadata = {
                        cell_name=XXX,
                        filename=XXX,
                        resource_id=XXX,
                        etype=XXX,
                        species=XXX,
                        brain_region=XXX,
                        ecodes={},
                        other_metadata={},
                    }
            targets (list): define the efeatures to extract as well as which
                protocols and current amplitude they should be extracted for. Of
                the form:

                .. code-block::

                    [{
                        "efeature_name": "AP_amplitude_1",
                        "efeature": "AP_amplitude",
                        "protocol": "IDRest",
                        "amplitude": 150.,
                        "tolerance": 10.,
                        "weight": 1.0,  # optional
                        "efel_settings": {
                            'stim_start': 200.,
                            'stim_end': 500.,
                            'Threshold': -10.
                        }
                    }]
            protocols_rheobase (list): names of the protocols that will be
                used to compute the rheobase of the cells. E.g: ['IDthresh'].
            available_traces (list of TraceFile)
            available_efeatures (llist of strings)
            auto_targets (list): if targets is not given, auto_targets
                define the efeatures to extract as well as which
                protocols and current amplitude they should be extracted for,
                given a list of possible protocols and amplitudes.
                min_recordings_per_amplitude, preferred_number_protocols and
                tolerance are optional.
                Of the form:

                .. code-block::

                    [{
                        "protocols": ["IDRest", "IV"],
                        "amplitudes": [150, 250],
                        "efeatures": ["AP_amplitude", "mean_frequency"],
                        "min_recordings_per_amplitude": 10,
                        "preferred_number_protocols": 1,
                        "tolerance": 10.,
                    }]
            additional_fitness_efeatures (list of dicts): efeatures to add to
                the output of the extraction, i.e. to the FitnessCalculatorConfiguration (FCC).
                These efeatures will not be extracted from the targets,
                but will be used during optimisation and / or validation.
                They should have the same format as the efeatures of the FCC, e.g.

                .. code-block::

                    [
                        {
                            "efel_feature_name": "Spikecount",
                            "protocol_name": "IDrest_130",
                            "recording_name": "soma.v",
                            "mean": 6.026,
                            "original_std": 4.016,
                            "efeature_name": "Spikecount",
                            "efel_settings": {
                                "strict_stiminterval": true,
                                "Threshold": -30.0,
                                "interp_step": 0.025,
                                "stim_start": 700.0,
                                "stim_end": 2700.0
                            }
                        }
                    ]
            additional_fitness_protocols (list of dicts): protocols to add to
                the output of the extraction, i.e. to the FitnessCalculatorConfiguration (FCC).
                These protocols will not be used during extraction,
                but will be used during optimisation and / or validation.
                They should have the same format as the protocols of the FCC, e.g.

                .. code-block::

                    [
                        {
                            "name": "IDrest_130",
                            "stimuli": [
                                {
                                "delay": 663.3473451327434,
                                "amp": 0.3491827776582547,
                                "thresh_perc": 130.56103157894464,
                                "duration": 2053.050884955752,
                                "totduration": 3000.0,
                                "holding_current": -0.13307330411636328
                                }
                            ],
                            "recordings_from_config": [
                                {
                                "type": "CompRecording",
                                "name": "IDrest_130.soma.v",
                                "location": "soma",
                                "variable": "v"
                                }
                            ],
                            "validation": false,
                            "protocol_type": "ThresholdBasedProtocol",
                            "stochasticity": false
                        }
                    ]
             protocols_mapping (dict, optional): maps original protocol
                identifiers to renamed versions for standardization.
                Defaults to None if not provided.

                Example:

                .. code-block::

                    {
                        "IDRest_150": 'Step_150_hyp',
                        "IDRest_200": 'Step_200_hyp',
                        "IDRest_250": 'Step_250_hyp',
                        "IV_-120": 'IV_-120_hyp',
                    }
        """

        self.available_traces = available_traces
        self.available_efeatures = available_efeatures

        self.files = []
        if files is not None:
            for f in files:
                tmp_trace = TraceFile(**f)
                if not self.is_trace_available(tmp_trace):
                    raise ValueError(f"File named {tmp_trace.cell_name} is not present on Nexus")
                self.files.append(tmp_trace)

        self.targets = []
        self.auto_targets = None
        if targets is not None:
            for t in targets:
                tmp_target = Target(**t)
                if not self.is_target_available(tmp_target):
                    raise ValueError(f"Efeature name {tmp_target.efeature} does not exist")
                self.targets.append(tmp_target)
        if auto_targets is not None:
            self.auto_targets = auto_targets

        if protocols_rheobase is None:
            self.protocols_rheobase = []
        elif isinstance(protocols_rheobase, str):
            self.protocols_rheobase = [protocols_rheobase]
        else:
            self.protocols_rheobase = protocols_rheobase

        self.additional_fitness_efeatures = additional_fitness_efeatures
        self.additional_fitness_protocols = additional_fitness_protocols

        self.protocols_mapping = protocols_mapping

    def is_trace_available(self, trace):
        if self.available_traces:
            available = next((a for a in self.available_traces if a == trace), False)
            return bool(available)
        return True

    def is_target_available(self, target):
        if self.available_efeatures and target.efeature not in self.available_efeatures:
            return False
        return True

    @property
    def files_metadata_BPE(self):
        """In BPE2 input format"""

        if self.files:
            files = self.files
        else:
            logger.info("No files given. Will use all available traces instead.")
            files = self.available_traces
        files_metadata = {}

        if self.targets:
            used_protocols = set([t.protocol for t in self.targets] + self.protocols_rheobase)
        elif self.auto_targets:
            used_protocols = set(
                [p for t in self.auto_targets for p in t["protocols"]] + self.protocols_rheobase
            )
        else:
            raise TypeError("either targets or autotargets should be set.")

        for f in files:
            for protocol in f.ecodes:
                if protocol in used_protocols:
                    if f.cell_name not in files_metadata:
                        files_metadata[f.cell_name] = {}
                    if protocol not in files_metadata[f.cell_name]:
                        files_metadata[f.cell_name][protocol] = []

                    ecodes_metadata = {
                        **f.ecodes.get(protocol, {}),
                        **f.other_metadata,
                        "filepath": f.filepath,
                    }

                    if "protocol_name" not in ecodes_metadata:
                        ecodes_metadata["protocol_name"] = protocol

                    files_metadata[f.cell_name][protocol].append(ecodes_metadata)

        for cell_name, protocols in files_metadata.items():
            if self.protocols_rheobase:
                for protocol in self.protocols_rheobase:
                    if protocol in protocols:
                        break
                else:
                    raise ValueError(
                        f"{protocol} is part of the protocols_rheobase but it has"
                        f" no associated ephys data for cell {cell_name}"
                    )
        return files_metadata

    @property
    def targets_BPE(self):
        """In BPE2 input format"""
        if not self.targets:
            return None
        return [t.as_dict() for t in self.targets]

    @property
    def auto_targets_BPE(self):
        """In BPE2 input format"""
        if not self.auto_targets:
            return None
        return [AutoTarget(**at) for at in self.auto_targets]

    @property
    def protocols_rheobase_BPE(self):
        """Returns None if empty"""
        if not self.protocols_rheobase:
            return None
        return self.protocols_rheobase

    @property
    def is_configuration_valid(self):
        """Checks that the configuration has targets, traces and that the targets can
        be found in the traces. This check can only be performed if the ecodes present
        in each files are known."""

        if not self.auto_targets:
            if not self.targets or not self.files:
                return False

        if self.targets and self.auto_targets:
            return False

        ecodes = set(
            itertools.chain(*[file.ecodes for file in self.files if file.ecodes is not None])
        )

        for target in self.targets:
            if ecodes and target.protocol not in ecodes:
                return False

        if self.auto_targets:
            for at in self.auto_targets:
                if not (
                    "protocols" in at.keys()
                    and "amplitudes" in at.keys()
                    and "efeatures" in at.keys()
                ):
                    return False

        return True

    def check_presence_RMP_Rin_efeatures(self, name_rmp_protocol, name_Rin_protocol):
        """Check that the protocols supposed to be used for RMP and Rin are present in the target
        and that they have the correct efeatures. If some features are missing, add them.
        """

        if self.targets:
            efeatures_rmp = [
                t.efeature
                for t in self.targets
                if are_same_protocol([t.protocol, t.amplitude], name_rmp_protocol)
            ]
            efeatures_rin = [
                t.efeature
                for t in self.targets
                if are_same_protocol([t.protocol, t.amplitude], name_Rin_protocol)
            ]
        elif self.auto_targets:
            efeatures_rmp = [
                efeat
                for t in self.auto_targets
                for efeat in t["efeatures"]
                if format_protocol_name_to_list(name_rmp_protocol)[0] in t["protocols"]
            ]
            efeatures_rin = [
                efeat
                for t in self.auto_targets
                for efeat in t["efeatures"]
                if format_protocol_name_to_list(name_Rin_protocol)[0] in t["protocols"]
            ]
        else:
            raise TypeError("either targets or autotargets should be set.")

        error_message = (
            "Target for feature {} is missing for protocol {}. Please add "
            "it if you wish to do a threshold-based optimisation."
        )

        if "voltage_base" not in efeatures_rmp:
            raise ValueError(error_message.format("voltage_base", name_rmp_protocol))
        if "voltage_base" not in efeatures_rin:
            raise ValueError(error_message.format("voltage_base", name_Rin_protocol))
        if "ohmic_input_resistance_vb_ssse" not in efeatures_rin:
            raise ValueError(
                error_message.format("ohmic_input_resistance_vb_ssse", name_Rin_protocol)
            )

    def get_related_nexus_ids(self):
        uses = []
        for f in self.files:
            if f.id:
                f_dict = {"id": f.id, "type": "Trace"}
                if f_dict not in uses:
                    uses.append(f_dict)

        return {"uses": uses}

    def as_dict(self):
        return {
            "files": [f.as_dict() for f in self.files],
            "targets": [t.as_dict() for t in self.targets],
            "protocols_rheobase": self.protocols_rheobase,
            "auto_targets": self.auto_targets,
            "additional_fitness_efeatures": self.additional_fitness_efeatures,
            "additional_fitness_protocols": self.additional_fitness_protocols,
            "protocols_mapping": self.protocols_mapping,
        }



"""ProtocolConfiguration"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import copy


class ProtocolConfiguration:
    """Container for the definition of a protocol"""

    def __init__(
        self,
        name,
        stimuli,
        recordings_from_config=None,
        recordings=None,
        validation=False,
        ion_variables=None,
        protocol_type="ThresholdBasedProtocol",
        stochasticity=False,
    ):
        """Init.

        The arguments efeatures and protocols are expected to be in the format used for the
        storage of the fitness calculator configuration. To store the results of an extraction,
        use the method init_from_bluepyefe.

        Args:
            name (str): name of the protocol
            stimuli (list of dict): contains the description of the stimuli. The exact format has
                to match what is expected by the related eCode class (see the classes defined
                in bluepyemodel.ecodes for more information). For example, for a Step protocol,
                the format will be:

                .. code-block::

                    [{
                        'amp': float, 'thresh_perc': float, 'holding_current': float,
                        'delay': float, 'duration': float, 'totduration': float, 'location': dict
                    }]
                    The location should have a format that can be read by
                    evaluation.evaluator.define_location()

            recordings_from_config (list of dict): contains the description of the recordings.
                For a recording at a given compartment, the format is for example:

                .. code-block::

                    [{
                        "type": "CompRecording",
                        "name": f"{protocol_name}.soma.v",
                        "location": "soma",
                        "variable": "v",
                    }]
            recordings (list of dict): same as recordings_from_config. Is here for backward
                compatibility only.
            ion_variables (list of str): ion current names and ionic concentration names
                for all available mechanisms.
            protocol_type (str): type of the protocol. Can be "ThresholdBasedProtocol" or
                "Protocol". When using "ThresholdBasedProtocol", the current amplitude and step
                amplitude of the stimulus will be ignored and replaced by values obtained from
                the holding current and rheobase of the cell model respectively. When using
                "Protocol", the current amplitude and step amplitude of the stimulus will be
                used directly, in this case, if a "thresh_perc" was informed, it will be ignored.
            stochasticity (bool): whether the mechanisms should be on stochastic mode
                when the protocol runs, or not.
        """

        self.name = name

        self.stimuli = stimuli
        if isinstance(self.stimuli, dict):
            self.stimuli = [self.stimuli]

        if recordings_from_config is None:
            if recordings is None:
                raise ValueError("Expected recordings_from_config to be not None")
            recordings_from_config = recordings
        if isinstance(recordings_from_config, dict):
            recordings_from_config = [recordings_from_config]

        self.recordings, self.recordings_from_config = self.init_recordings(
            recordings_from_config, ion_variables
        )

        self.validation = validation
        self.protocol_type = protocol_type

        self.stochasticity = stochasticity

    def init_recordings(self, recordings_from_config=None, ion_variables=None):
        new_recordings = []
        new_recordings_from_config = []
        for recording in recordings_from_config:
            new_recordings.append(recording)
            new_recordings_from_config.append(recording)

            if ion_variables is not None:
                for ion in ion_variables:
                    new_rec = recording.copy()

                    if "variable" in recording:
                        new_rec["variable"] = ion
                    elif "var" in recording:
                        new_rec["var"] = ion
                    else:
                        raise KeyError("Expected 'var' or 'variable' in recording list.")

                    new_rec["name"] = ".".join(new_rec["name"].split(".")[:-1] + [ion])

                    new_recordings.append(new_rec)

        return new_recordings, new_recordings_from_config

    def as_dict(self):
        """Dictionary form"""

        prot_as_dict = copy.deepcopy(vars(self))
        prot_as_dict.pop("recordings")
        return prot_as_dict



"""Function related to parallel computing"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import datetime
import multiprocessing
import os
from multiprocessing import pool

from bluepyemodel.tools.utils import logger


def ipyparallel_map_function(os_env_profile="IPYTHON_PROFILE", profile=None):
    """Get the map function linked to the ipython profile

    Args:
        os_env_profile (str): The name of the environment variable
            that holds the profile name for ipyparallel.
            If this environment variable exists and contains a valid profile name,
            it will be used to create an ipyparallel Client. Defaults to "IPYTHON_PROFILE"
        profile (str): name of the ipython profile

    Returns:
        map
    """
    if not profile:
        profile = os.getenv(os_env_profile)
    if profile:
        from ipyparallel import Client

        rc = Client(profile=profile)
        lview = rc.load_balanced_view()

        def mapper(func, it):
            start_time = datetime.datetime.now()
            ret = lview.map_sync(func, it)
            logger.debug("Took %s", datetime.datetime.now() - start_time)
            return ret

    else:
        logger.warning(
            "Environment variable 'os_env_profile' not set or invalid;"
            "falling back to the default map function."
        )
        mapper = map

    return mapper


class NoDaemonProcess(multiprocessing.Process):
    """Class that represents a non-daemon process"""

    # pylint: disable=dangerous-default-value

    def __init__(self, group=None, target=None, name=None, args=(), kwargs={}):
        """Ensures group=None, for macosx."""
        super().__init__(group=None, target=target, name=name, args=args, kwargs=kwargs)

    def _get_daemon(self):
        """Get daemon flag"""
        return False

    def _set_daemon(self, value):
        """Set daemon flag"""

    daemon = property(_get_daemon, _set_daemon)


class NestedPool(pool.Pool):  # pylint: disable=abstract-method
    """Class that represents a MultiProcessing nested pool"""

    Process = NoDaemonProcess


def get_mapper(backend, ipyparallel_profile=None):
    """Get a mapper for parallel computations."""
    if backend == "ipyparallel":
        return ipyparallel_map_function(profile=ipyparallel_profile)

    if backend == "multiprocessing":
        nested_pool = NestedPool()
        return nested_pool.map

    # For any other backend, default to the built-in map function
    return map



"""Module with protocol classes."""

"""
Copyright 2023-2024, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging
from collections import OrderedDict

import numpy
from bluepyopt import ephys

from ..ecode import eCodes
from .recordings import LooseDtRecordingCustom
from .recordings import check_recordings

# pylint: disable=abstract-method

logger = logging.getLogger(__name__)


class BPEMProtocol(ephys.protocols.SweepProtocol):
    """Base protocol"""

    def __init__(
        self,
        name=None,
        stimulus=None,
        recordings=None,
        cvode_active=None,
        stochasticity=False,
    ):
        """Constructor

        Args:
            name (str): name of this object
            stimulus (Stimulus): stimulus objects
            recordings (list of Recordings): Recording objects used in the
                protocol
            cvode_active (bool): whether to use variable time step
            stochasticity (bool): turns on or off the channels that can be
                stochastic
        """

        super().__init__(
            name=name,
            stimuli=[stimulus],
            recordings=recordings,
            cvode_active=cvode_active,
            deterministic=not stochasticity,
        )

        self.stimulus = stimulus

        self.features = []

    def instantiate(self, sim=None, cell_model=None):
        """Check recordings, then instantiate."""
        if not all(rec.checked for rec in self.recordings):
            self.recordings = check_recordings(self.recordings, cell_model.icell, sim)

        super().instantiate(sim, cell_model)

    def stim_start(self):
        """Time stimulus starts"""
        return self.stimulus.stim_start

    def stim_end(self):
        return self.stimulus.stim_end

    def amplitude(self):
        return self.stimulus.amplitude

    def run(  # pylint: disable=arguments-differ, arguments-renamed
        self,
        cell_model,
        param_values=None,
        sim=None,
        isolate=None,
        timeout=None,
        responses=None,
    ):
        """Run protocol"""

        param_values = {} if param_values is None else param_values
        responses = {} if responses is None else responses

        return super().run(
            cell_model=cell_model,
            param_values=param_values,
            sim=sim,
            isolate=isolate,
            timeout=timeout,
        )


class ResponseDependencies:
    """To add to a protocol to specify that it depends on the responses of other protocols"""

    def __init__(self, dependencies=None):
        """Constructor

        Args:
            dependencies (dict): dictionary of dependencies of the form
                {self_attribute_name: [protocol_name, response_name]}.
        """

        self.dependencies = {} if dependencies is None else dependencies

    def return_none_responses(self):
        # pylint: disable=inconsistent-return-statements
        raise NotImplementedError()

    def set_attribute(self, attribute, value):
        """Set an attribute of the class based on the name of the attribute. Also handles
        the case where the name is of the form: attribute.sub_attribute"""

        if "." in attribute:
            obj2 = getattr(self, attribute.split(".")[0])
            setattr(obj2, attribute.split(".")[1], value)
        else:
            setattr(self, attribute, value)

    def set_dependencies(self, responses):
        for attribute_name, dep in self.dependencies.items():
            if responses.get(dep[1], None) is None:
                logger.debug("Dependency %s missing", dep[1])
                return False
            self.set_attribute(attribute_name, responses[dep[1]])
        return True

    def _run(
        self,
        cell_model,
        param_values=None,
        sim=None,
        isolate=None,
        timeout=None,
        responses=None,
    ):
        raise NotImplementedError("The run code of the sub-classes goes here!")

    def run(
        self,
        cell_model,
        param_values=None,
        sim=None,
        isolate=None,
        timeout=None,
        responses=None,
    ):
        """Run protocol"""

        param_values = {} if param_values is None else param_values
        responses = {} if responses is None else responses

        if not self.set_dependencies(responses):
            return self.return_none_responses()

        return self._run(cell_model, param_values, sim, isolate, timeout)


class ProtocolWithDependencies(BPEMProtocol, ResponseDependencies):
    """To add to a protocol to specify that it depends on the responses of other protocols"""

    def __init__(
        self,
        dependencies=None,
        name=None,
        stimulus=None,
        recordings=None,
        cvode_active=None,
        stochasticity=False,
    ):
        """Constructor

        Args:
            dependencies (dict): dictionary of dependencies of the form
                {self_attribute_name: [protocol_name, response_name]}.
        """

        ResponseDependencies.__init__(self, dependencies)
        BPEMProtocol.__init__(
            self,
            name=name,
            stimulus=stimulus,
            recordings=recordings,
            cvode_active=cvode_active,
            stochasticity=stochasticity,
        )

    def _run(
        self,
        cell_model,
        param_values=None,
        sim=None,
        isolate=None,
        timeout=None,
        responses=None,
    ):
        return BPEMProtocol.run(self, cell_model, param_values, sim, isolate, timeout, responses)


class ThresholdBasedProtocol(ProtocolWithDependencies):
    """Protocol having rheobase-rescaling capabilities. When using ThresholdBasedProtocol,
    the holding current amplitude and step amplitude of the stimulus will be ignored and
    replaced by values obtained from the holding current and rheobase of the cell model
    respectively."""

    def __init__(
        self,
        name=None,
        stimulus=None,
        recordings=None,
        cvode_active=None,
        stochasticity=False,
        hold_key="bpo_holding_current",
        thres_key="bpo_threshold_current",
        hold_prot_name="SearchHoldingCurrent",
        thres_prot_name="SearchThresholdCurrent",
    ):
        """Constructor"""

        dependencies = {
            "stimulus.holding_current": [hold_prot_name, hold_key],
            "stimulus.threshold_current": [thres_prot_name, thres_key],
        }

        ProtocolWithDependencies.__init__(
            self,
            dependencies=dependencies,
            name=name,
            stimulus=stimulus,
            recordings=recordings,
            cvode_active=cvode_active,
            stochasticity=stochasticity,
        )

    def return_none_responses(self):
        return {k.name: None for k in self.recordings}

    def run(
        self,
        cell_model,
        param_values=None,
        sim=None,
        isolate=None,
        timeout=None,
        responses=None,
    ):
        return ResponseDependencies.run(
            self, cell_model, param_values, sim, isolate, timeout, responses
        )


class RMPProtocol(BPEMProtocol):
    """Protocol consisting of a step of amplitude zero"""

    def __init__(
        self,
        name,
        location,
        target_voltage,
        stimulus_duration=500.0,
        output_key="bpo_rmp",
    ):
        """Constructor"""

        stimulus_definition = {
            "delay": 0.0,
            "amp": 0.0,
            "thresh_perc": None,
            "duration": stimulus_duration,
            "totduration": stimulus_duration,
            "holding_current": 0.0,
        }

        self.recording_name = f"{name}.{location.name}.v"
        stimulus = eCodes["step"](location=location, **stimulus_definition)
        recordings = [
            LooseDtRecordingCustom(name=self.recording_name, location=location, variable="v")
        ]
        self.output_key = output_key

        BPEMProtocol.__init__(
            self,
            name=name,
            stimulus=stimulus,
            recordings=recordings,
            cvode_active=True,
            stochasticity=False,
        )

        self.stimulus_duration = stimulus_duration

        self.target_voltage = target_voltage
        self.target_voltage.stim_start = stimulus_duration - 100.0
        self.target_voltage.stim_end = stimulus_duration
        self.target_voltage.stimulus_current = 0.0

    def run(
        self,
        cell_model,
        param_values=None,
        sim=None,
        isolate=None,
        timeout=None,
        responses=None,
    ):
        """Compute the RMP"""

        response = BPEMProtocol.run(
            self,
            cell_model,
            param_values,
            sim=sim,
            isolate=isolate,
            timeout=timeout,
            responses=responses,
        )
        if responses is None or response[self.recording_name] is None:
            return {self.recording_name: None, self.output_key: None}

        bpo_rmp = self.target_voltage.calculate_feature(response)
        response[self.output_key] = bpo_rmp if bpo_rmp is None else bpo_rmp[0]

        return response


class RinProtocol(ProtocolWithDependencies):
    """Protocol used to find the input resistance of a model"""

    def __init__(
        self,
        name,
        location,
        target_rin,
        amp=-0.02,
        stimulus_delay=500.0,
        stimulus_duration=500.0,
        totduration=1000.0,
        output_key="bpo_rin",
        hold_key="bpo_holding_current",
        hold_prot_name="SearchHoldingCurrent",
    ):
        """Constructor"""

        stimulus_definition = {
            "delay": stimulus_delay,
            "amp": amp,
            "thresh_perc": None,
            "duration": stimulus_duration,
            "totduration": totduration,
            "holding_current": None,
        }

        self.recording_name = f"{name}.{location.name}.v"
        stimulus = eCodes["step"](location=location, **stimulus_definition)
        recordings = [
            LooseDtRecordingCustom(name=self.recording_name, location=location, variable="v")
        ]
        self.output_key = output_key

        dependencies = {"stimulus.holding_current": [hold_prot_name, hold_key]}

        ProtocolWithDependencies.__init__(
            self,
            dependencies=dependencies,
            name=name,
            stimulus=stimulus,
            recordings=recordings,
            cvode_active=True,
            stochasticity=False,
        )

        self.target_rin = target_rin
        self.target_rin.stim_start = stimulus_delay
        self.target_rin.stim_end = stimulus_delay + stimulus_duration
        self.target_rin.stimulus_current = amp

    def return_none_responses(self):
        return {self.recording_name: None, self.output_key: None}

    def run(
        self,
        cell_model,
        param_values=None,
        sim=None,
        isolate=None,
        timeout=None,
        responses=None,
    ):
        """Compute the Rin"""

        response = ResponseDependencies.run(
            self,
            cell_model,
            param_values,
            sim=sim,
            isolate=isolate,
            timeout=timeout,
            responses=responses,
        )

        bpo_rin = self.target_rin.calculate_feature(response)
        response[self.output_key] = bpo_rin if bpo_rin is None else bpo_rin[0]

        return response


class NoHoldingCurrent(ephys.protocols.Protocol):
    """Empty class returning a holding current of zero."""

    def __init__(self, name, output_key="bpo_holding_current"):
        """Constructor."""
        super().__init__(
            name=name,
        )
        self.output_key = output_key
        self.recordings = {}

    def run(
        self,
        cell_model,
        param_values=None,
        sim=None,
        isolate=None,
        timeout=None,
        responses=None,
    ):
        # pylint: disable=unused-argument
        return {self.output_key: 0}


class SearchHoldingCurrent(BPEMProtocol):
    """Protocol used to find the holding current of a model"""

    def __init__(
        self,
        name,
        location,
        target_voltage=None,
        voltage_precision=0.1,
        stimulus_duration=500.0,
        upper_bound=0.2,
        lower_bound=-0.2,
        strict_bounds=True,
        max_depth=7,
        no_spikes=True,
        output_key="bpo_holding_current",
    ):
        """Constructor

        Args:
            name (str): name of this object
            location (Location): location on which to perform the search (
                usually the soma).
            target_voltage (EFeature): target for the voltage at holding_current
            voltage_precision (float): accuracy for holding voltage, in mV, to stop the search
            stimulus_duration (float): length of the protocol
            upper_bound (float): upper bound for the holding current, in pA
            lower_bound (float): lower bound for the holding current, in pA
            strict_bounds (bool): to adaptively enlarge bounds if current is outside
            max_depth (int): maximum depth for the binary search
            no_spikes (bool): if True, the holding current will only be considered valid if there
                are no spikes at holding.
        """

        stimulus_definition = {
            "delay": 0.0,
            "amp": 0.0,
            "thresh_perc": None,
            "duration": stimulus_duration,
            "totduration": stimulus_duration,
            "holding_current": 0.0,
        }

        self.recording_name = f"{name}.{location.name}.v"
        stimulus = eCodes["step"](location=location, **stimulus_definition)
        recordings = [
            LooseDtRecordingCustom(name=self.recording_name, location=location, variable="v")
        ]
        self.output_key = output_key

        BPEMProtocol.__init__(
            self,
            name=name,
            stimulus=stimulus,
            recordings=recordings,
            cvode_active=True,
            stochasticity=False,
        )

        self.voltage_precision = voltage_precision
        self.upper_bound = upper_bound
        self.lower_bound = lower_bound
        self.strict_bounds = strict_bounds
        self.no_spikes = no_spikes

        self.target_voltage = target_voltage
        self.holding_voltage = self.target_voltage.exp_mean

        self.target_voltage.stim_start = stimulus_duration - 100.0
        self.target_voltage.stim_end = stimulus_duration
        self.target_voltage.stimulus_current = 0.0

        self.max_depth = max_depth

        self.spike_feature = ephys.efeatures.eFELFeature(
            name="SearchHoldingCurrent.Spikecount",
            efel_feature_name="Spikecount",
            recording_names={"": f"SearchHoldingCurrent.{location.name}.v"},
            stim_start=0.0,
            stim_end=stimulus_duration,
            exp_mean=1,
            exp_std=0.1,
        )

    def get_voltage_base(
        self, holding_current, cell_model, param_values, sim, isolate, timeout=None
    ):
        """Calculate voltage base for a certain holding current"""

        self.stimuli[0].amp = holding_current
        response = BPEMProtocol.run(
            self, cell_model, param_values, sim=sim, isolate=isolate, timeout=timeout
        )
        if response is None or response[self.recording_name] is None:
            return None

        if self.no_spikes:
            n_spikes = self.spike_feature.calculate_feature(response)
            if n_spikes is None or n_spikes > 0:
                return None

        voltage_base = self.target_voltage.calculate_feature(response)
        if voltage_base is None:
            return None
        return voltage_base[0]

    def run(
        self,
        cell_model,
        param_values=None,
        sim=None,
        isolate=None,
        timeout=None,
        responses=None,
    ):
        """Run protocol"""
        if not self.strict_bounds:
            # first readjust the bounds if needed
            voltage_min = 1e10
            while voltage_min > self.target_voltage.exp_mean:
                voltage_min = self.get_voltage_base(
                    holding_current=self.lower_bound,
                    cell_model=cell_model,
                    param_values=param_values,
                    sim=sim,
                    isolate=isolate,
                    timeout=timeout,
                )
                if voltage_min is None:
                    voltage_min = 1e10

                if voltage_min > self.target_voltage.exp_mean:
                    self.lower_bound -= 0.2

            voltage_max = -1e10
            while voltage_max < self.target_voltage.exp_mean:
                voltage_max = self.get_voltage_base(
                    holding_current=self.upper_bound,
                    cell_model=cell_model,
                    param_values=param_values,
                    sim=sim,
                    isolate=isolate,
                    timeout=timeout,
                )
                if voltage_max is None:
                    # if we spike, we let it pass to the search
                    voltage_max = 1e10

                elif voltage_max < self.target_voltage.exp_mean:
                    self.upper_bound += 0.2

        response = {
            self.output_key: self.bisection_search(
                cell_model,
                param_values,
                sim=sim,
                isolate=isolate,
                upper_bound=self.upper_bound,
                lower_bound=self.lower_bound,
                timeout=timeout,
            )
        }

        if response[self.output_key] is None:
            return response

        response.update(
            BPEMProtocol.run(
                self,
                cell_model,
                param_values,
                sim=sim,
                isolate=isolate,
                timeout=timeout,
            )
        )

        return response

    def bisection_search(
        self,
        cell_model,
        param_values,
        sim,
        isolate,
        upper_bound,
        lower_bound,
        timeout=None,
        depth=0,
    ):
        """Do bisection search to find holding current"""
        mid_bound = (upper_bound + lower_bound) * 0.5
        voltage = self.get_voltage_base(
            holding_current=mid_bound,
            cell_model=cell_model,
            param_values=param_values,
            sim=sim,
            isolate=isolate,
            timeout=timeout,
        )
        # if we don't converge fast enough, we stop and return lower bound, which will not spike
        if depth > self.max_depth:
            logging.debug(
                "Exiting search due to reaching max_depth. The required voltage precision "
                "was not reached."
            )
            return lower_bound

        if voltage is not None and abs(voltage - self.holding_voltage) < self.voltage_precision:
            logger.debug("Depth of holding search: %s", depth)
            return mid_bound

        # if voltage is None, it means we spike at mid_bound, so we try with lower side
        if voltage is None or voltage > self.holding_voltage:
            return self.bisection_search(
                cell_model,
                param_values,
                sim=sim,
                isolate=isolate,
                lower_bound=lower_bound,
                upper_bound=mid_bound,
                depth=depth + 1,
            )

        return self.bisection_search(
            cell_model,
            param_values,
            sim=sim,
            isolate=isolate,
            lower_bound=mid_bound,
            upper_bound=upper_bound,
            depth=depth + 1,
        )


class SearchThresholdCurrent(ProtocolWithDependencies):
    """Protocol used to find the threshold current (rheobase) of a model"""

    def __init__(
        self,
        name,
        location,
        target_threshold=None,
        current_precision=1e-2,
        stimulus_delay=500.0,
        stimulus_duration=2000.0,
        stimulus_totduration=3000.0,
        max_threshold_voltage=-30,
        spikecount_timeout=50,
        max_depth=10,
        no_spikes=True,
        efel_threshold=None,
        output_key="bpo_threshold_current",
        hold_key="bpo_holding_current",
        rmp_key="bpo_rmp",
        rin_key="bpo_rin",
        hold_prot_name="SearchHoldingCurrent",
        rmp_prot_name="RMPProtocol",
        rin_prot_name="RinProtocol",
    ):
        """Constructor.

        Args:
            name (str): name of this object
            location (Location): location on which to perform the search (
                usually the soma).
            target_threshold (Efeature): target for the threshold_current
            current_precision (float): size of search interval in current to stop the search
            stimulus_delay (float): delay before the beginning of the step
                used to create the protocol
            stimulus_duration (float): duration of the step used to create the
                protocol
            stimulus_totduration (float): total duration of the protocol
            max_threshold_voltage (float): maximum voltage used as upper
                bound in the threshold current search
            spikecount_timeout (float): timeout for spikecount computation, if timeout is reached,
                we set spikecount=2 as if many spikes were present, to speed up bisection search.
            max_depth (int): maximum depth for the binary search
            no_spikes (bool): if True, will check that the holding current (lower bound) does not
                trigger spikes.
            efel_threshold: spike threshold for the efel settings.
                Set to None to keep the default value (currently -20 mV in efel)
        """
        # pylint: disable=too-many-arguments

        dependencies = {
            "stimulus.holding_current": [hold_prot_name, hold_key],
            "rin": [rin_prot_name, rin_key],
            "rmp": [rmp_prot_name, rmp_key],
        }

        stimulus_definition = {
            "delay": stimulus_delay,
            "amp": 0.0,
            "thresh_perc": None,
            "duration": stimulus_duration,
            "totduration": stimulus_totduration,
            "holding_current": None,
        }

        self.recording_name = f"{name}.{location.name}.v"
        stimulus = eCodes["step"](location=location, **stimulus_definition)
        recordings = [
            LooseDtRecordingCustom(name=self.recording_name, location=location, variable="v")
        ]
        self.output_key = output_key
        self.hold_key = hold_key

        super().__init__(
            dependencies=dependencies,
            name=name,
            stimulus=stimulus,
            recordings=recordings,
            cvode_active=True,
            stochasticity=False,
        )

        self.rin = None
        self.rmp = None

        self.target_threshold = target_threshold
        self.max_threshold_voltage = max_threshold_voltage
        self.current_precision = current_precision
        self.no_spikes = no_spikes
        self.max_depth = max_depth

        self.spike_feature = ephys.efeatures.eFELFeature(
            name=f"{name}.Spikecount",
            efel_feature_name="Spikecount",
            recording_names={"": f"{name}.{location.name}.v"},
            stim_start=stimulus_delay,
            stim_end=stimulus_delay + stimulus_duration,
            exp_mean=1,
            exp_std=0.1,
            threshold=efel_threshold,
        )
        self.spikecount_timeout = spikecount_timeout

    def return_none_responses(self):
        return {self.output_key: None, self.recording_name: None}

    def _get_spikecount(self, current, cell_model, param_values, sim, isolate):
        """Get spikecount at a given current."""

        self.stimulus.amp = current

        response = self._run(
            cell_model,
            param_values,
            sim=sim,
            isolate=isolate,
            timeout=self.spikecount_timeout,
        )
        if response[self.recording_name] is None:
            logger.debug(
                "Trace computation for threshold timed out at %s",
                self.spikecount_timeout,
            )
            return 2

        return self.spike_feature.calculate_feature(response)

    def run(
        self,
        cell_model,
        param_values=None,
        sim=None,
        isolate=None,
        timeout=None,
        responses=None,
    ):
        """Run protocol"""
        if not self.set_dependencies(responses):
            return self.return_none_responses()

        lower_bound, upper_bound = self.define_search_bounds(
            cell_model, param_values, sim, isolate, responses
        )
        if lower_bound is None or upper_bound is None:
            logger.debug("Threshold search bounds are not good")
            return {self.output_key: None}

        threshold = self.bisection_search(
            cell_model,
            param_values,
            sim,
            isolate,
            upper_bound=upper_bound,
            lower_bound=lower_bound,
            timeout=timeout,
        )

        response = {self.output_key: threshold}
        if threshold is None:
            return response

        self.stimulus.amp = threshold
        response.update(
            self._run(cell_model, param_values, sim=sim, isolate=isolate, timeout=timeout)
        )

        return response

    def max_threshold_current(self):
        """Find the current necessary to get to max_threshold_voltage"""
        max_threshold_current = (self.max_threshold_voltage - self.rmp) / self.rin
        max_threshold_current = numpy.min([max_threshold_current, 2.0])
        logger.debug("Max threshold current: %.6g", max_threshold_current)
        return max_threshold_current

    def define_search_bounds(self, cell_model, param_values, sim, isolate, responses):
        """Define the bounds and check their validity"""

        upper_bound = self.max_threshold_current()
        spikecount = self._get_spikecount(upper_bound, cell_model, param_values, sim, isolate)
        if spikecount == 0:
            logger.debug("No spikes at upper bound during threshold search")
            return None, None

        lower_bound = responses[self.hold_key]
        spikecount = self._get_spikecount(lower_bound, cell_model, param_values, sim, isolate)

        if spikecount > 0:
            if self.no_spikes:
                logger.debug("Spikes at lower bound during threshold search")
                return None, None
            lower_bound -= 0.5

        if lower_bound > upper_bound:
            logger.debug("lower bound higher than upper bound in threshold search")
            return lower_bound, lower_bound

        return lower_bound, upper_bound

    def bisection_search(
        self,
        cell_model,
        param_values,
        sim,
        isolate,
        upper_bound,
        lower_bound,
        timeout=None,
        depth=0,
    ):
        """Do bisection search to find threshold current."""
        mid_bound = (upper_bound + lower_bound) * 0.5
        spikecount = self._get_spikecount(mid_bound, cell_model, param_values, sim, isolate)
        if abs(lower_bound - upper_bound) < self.current_precision:
            logger.debug("Depth of threshold search: %s", depth)
            return upper_bound

        if depth > self.max_depth:
            return upper_bound

        if spikecount == 0:
            return self.bisection_search(
                cell_model,
                param_values,
                sim=sim,
                isolate=isolate,
                lower_bound=mid_bound,
                upper_bound=upper_bound,
                timeout=timeout,
                depth=depth + 1,
            )
        return self.bisection_search(
            cell_model,
            param_values,
            sim=sim,
            isolate=isolate,
            lower_bound=lower_bound,
            upper_bound=mid_bound,
            timeout=timeout,
            depth=depth + 1,
        )


class ProtocolRunner(ephys.protocols.Protocol):
    """Meta-protocol in charge of running the other protocols in the correct order"""

    def __init__(self, protocols, name="ProtocolRunner"):
        """Initialize the protocol runner

        Args:
            protocols (dict): Dictionary of protocols to run
            name (str): Name of the current protocol runner
        """

        super().__init__(name=name)

        self.protocols = protocols
        self.execution_order = self.compute_execution_order()

    def _add_to_execution_order(self, protocol, execution_order, before_index=None):
        """Recursively adds protocols to the execution order while making sure that their
        dependencies are added before them. Warning: Does not solve all types of dependency graph.
        """

        if protocol.name not in execution_order:
            if before_index is None:
                execution_order.append(protocol.name)
            else:
                execution_order.insert(before_index, protocol.name)

        if hasattr(protocol, "dependencies"):
            for dep in protocol.dependencies.values():
                if dep[0] not in execution_order:
                    self._add_to_execution_order(
                        self.protocols[dep[0]],
                        execution_order,
                        before_index=execution_order.index(protocol.name),
                    )

    def compute_execution_order(self):
        """Compute the execution order of the protocols by taking into account their dependencies"""

        execution_order = []

        for protocol in self.protocols.values():
            self._add_to_execution_order(protocol, execution_order)

        return execution_order

    def run(self, cell_model, param_values, sim=None, isolate=None, timeout=None):
        """Run protocol"""

        responses = OrderedDict()
        cell_model.freeze(param_values)

        for protocol_name in self.execution_order:
            logger.debug("Computing protocol %s", protocol_name)
            new_responses = self.protocols[protocol_name].run(
                cell_model,
                param_values={},
                sim=sim,
                isolate=isolate,
                timeout=timeout,
                responses=responses,
            )

            if new_responses is None or any(v is None for v in new_responses.values()):
                logger.debug("None in responses, exiting evaluation")
                break

            responses.update(new_responses)

        cell_model.unfreeze(param_values.keys())
        return responses

    def __str__(self):
        """String representation"""

        content = f"Sequence protocol {self.name}:\n"

        content += f"{len(self.protocols)} subprotocols:\n"
        for protocol in self.protocols:
            content += f"{protocol}\n"

        return content



"""DataAccessPoint class."""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import glob
import logging
import pathlib
import pickle
from enum import Enum
from itertools import chain

import efel
import numpy
from bluepyopt.deapext.algorithms import _check_stopping_criteria
from bluepyopt.deapext.stoppingCriteria import MaxNGen

from bluepyemodel.emodel_pipeline.emodel_metadata import EModelMetadata
from bluepyemodel.emodel_pipeline.emodel_settings import EModelPipelineSettings
from bluepyemodel.tools.utils import get_checkpoint_path
from bluepyemodel.tools.utils import read_checkpoint

# pylint: disable=no-member,unused-argument,assignment-from-no-return,no-value-for-parameter

logger = logging.getLogger(__name__)


class OptimisationState(Enum):
    """
    the possible states of an optimisation process.

    Attributes:
        COMPLETED (str): The optimisation process has completed successfully.
        IN_PROGRESS (str): The optimisation process is currently in progress.
        EMPTY (str): The optimisation process has not yet been started.
    """

    COMPLETED = "completed"
    IN_PROGRESS = "in_progress"
    EMPTY = "empty"


class DataAccessPoint:
    """Abstract data access point class. This class is not meant to be used directly.
    Instead, it is used through the classes LocalAccessPoint and NexusAccessPoint.
    """

    def __init__(
        self,
        emodel=None,
        etype=None,
        ttype=None,
        mtype=None,
        species=None,
        brain_region=None,
        iteration_tag=None,
        synapse_class=None,
    ):
        """Init"""

        self.emodel_metadata = EModelMetadata(
            emodel,
            etype,
            ttype,
            mtype,
            species,
            brain_region,
            iteration_tag,
            synapse_class,
        )

    def set_emodel(self, emodel):
        """Setter for the name of the emodel."""
        self.emodel_metadata.emodel = emodel

    def get_pipeline_settings(self):
        """ """
        return EModelPipelineSettings()

    def store_efeatures(
        self,
        efeatures,
        current,
    ):
        """Save the efeatures and currents obtained from BluePyEfe"""

    def store_protocols(self, stimuli):
        """Save the protocols obtained from BluePyEfe"""

    def store_emodel(
        self,
        scores,
        params,
        optimiser_name,
        seed,
        validated=None,
        scores_validation=None,
    ):
        """Save a model obtained from BluePyOpt"""

    def get_emodel(self):
        """Get dict with parameter of single emodel (including seed if any)"""

    def get_emodels(self, emodels):
        """Get the list of emodels dictionaries."""

    def store_targets_configuration(self):
        """Store the configuration of the targets (targets and ephys files used)"""

    def get_targets_configuration(self):
        """Get the configuration of the targets (targets and ephys files used)"""

    def store_model_configuration(self):
        """Store the configuration of a model, including parameters, mechanisms and distributions"""

    def get_available_mechanisms(self):
        """Get all the available mechanisms"""

    def get_available_efeatures(self, cleaned=True):
        """Returns a curated list of available eFEL features"""

        efel_features = efel.get_feature_names()

        if not cleaned:
            return efel_features

        features = []
        for f in efel_features:
            if f.endswith("indices"):
                continue
            if f in ["voltage", "time", "current"]:
                continue
            features.append(f)

        features.remove("AP_begin_time")
        features.remove("peak_time")

        return features

    def get_available_traces(self, filter_species=False, filter_brain_region=False):
        """Get the list of available Traces for the current species from Nexus"""
        return None

    def get_distributions(self):
        """Get the list of available distributions"""

    def store_distribution(self, distribution):
        """Store a channel distribution as a resource of type EModelChannelDistribution"""

    def get_model_configuration(self, skip_get_available_morph=True):
        """Get the configuration of the model, including parameters, mechanisms and distributions"""

    def store_fitness_calculator_configuration(self, configuration):
        """Store a fitness calculator configuration"""

    def get_fitness_calculator_configuration(self, record_ions_and_currents=False):
        """Get the configuration of the fitness calculator (efeatures and protocols)"""

    def get_mechanisms_directory(self):
        """Return the path to the directory containing the mechanisms for the current emodel"""

    def download_mechanisms(self):
        """Download the mod files when not already downloaded"""

    def get_emodel_names(self):
        """Get the list of all the names of emodels

        Returns:
            dict: keys are emodel names with seed, values are names without seed.
        """

    def store_hocs(
        self,
        only_validated=False,
        only_best=True,
        seeds=None,
        map_function=map,
        new_emodel_name=None,
        description=None,
        output_base_dir="export_emodels_hoc",
    ):
        """Store the hoc files"""

    def store_emodels_hoc(
        self,
        only_validated=False,
        only_best=True,
        seeds=None,
        map_function=map,
        new_emodel_name=None,
        description=None,
    ):
        """Store hoc file produced by export_hoc"""

    def store_emodels_sonata(
        self,
        only_validated=False,
        only_best=True,
        seeds=None,
        map_function=map,
        new_emodel_name=None,
        description=None,
    ):
        """Store hoc file produced by export_sonata"""

    def optimisation_state(self, seed=None, continue_opt=False):
        """Return the state of the optimisation.

        Args:
            seed (int): seed used in the optimisation.
            continue_opt (bool): whether to continue optimisation or not
                when the optimisation is not complete

        Raises:
            ValueError if optimiser in pipeline settings in neither
                "SO-CMA", "MO-CMA" or "IBEA"

        Returns:
            bool: True if completed, False if in progress or empty
        """

        checkpoint_path = get_checkpoint_path(self.emodel_metadata, seed=seed)

        # no file -> target not complete
        if not pathlib.Path(checkpoint_path).is_file():
            return OptimisationState.EMPTY

        # there is a file & continue opt is False -> target considered complete
        if not continue_opt:
            return OptimisationState.COMPLETED

        # there is a file & we want to continue optimisation -> check if optimisation if finished
        optimiser = self.pipeline_settings.optimiser
        ngen = self.pipeline_settings.max_ngen

        with open(str(checkpoint_path), "rb") as checkpoint_file:
            cp = pickle.load(checkpoint_file, encoding="latin1")

        # CMA
        if optimiser in ["SO-CMA", "MO-CMA"]:
            gen = cp["generation"]
            CMA_es = cp["CMA_es"]
            CMA_es.check_termination(gen)
            # no termination met -> still active -> target not complete
            if CMA_es.active:
                return OptimisationState.IN_PROGRESS
            return OptimisationState.COMPLETED

        # IBEA
        if optimiser == "IBEA":
            gen = cp["generation"]
            stopping_criteria = [MaxNGen(ngen)]
            stopping_params = {"gen": gen}
            run_complete = _check_stopping_criteria(stopping_criteria, stopping_params)
            if run_complete:
                return OptimisationState.COMPLETED
            return OptimisationState.IN_PROGRESS

        raise ValueError(f"Unknown optimiser: {optimiser}")

    def get_ion_currents_concentrations(self):
        """Get all ion currents and ion concentrations.

        Returns:
            tuple containing:

            (list of str): current (ion and non-specific) names for all available mechanisms
            (list of str): ionic concentration names for all available mechanisms
        """
        # pylint: disable=assignment-from-no-return
        mechs = self.get_available_mechanisms()
        if mechs is None:
            return None, None
        ion_currents = list(chain.from_iterable([mech.get_current() for mech in mechs]))
        ionic_concentrations = list(
            chain.from_iterable([mech.ionic_concentrations for mech in mechs])
        )
        # append i_pas which is present by default
        ion_currents.append("i_pas")
        return ion_currents, ionic_concentrations

    def __str__(self):
        str_ = "#############################################################\n"
        str_ += "################## SUMMARY: EMODEL CREATION #################\n"
        str_ += "#############################################################\n\n"

        str_ += "GENERAL INFORMATION\n"
        str_ += f"  Current working directory: {pathlib.Path.cwd()}\n"
        str_ += f"  Type of access point: {type(self).__name__}\n\n"

        str_ += "EMODEL METADATA\n"
        for k, v in vars(self.emodel_metadata).items():
            if v is not None:
                str_ += f"  {k}: {v}\n"
        str_ += "\n"

        str_ += "CONFIGURATION STATUS\n"
        str_ += f"  Has pipeline settings: {self.has_pipeline_settings()}\n"
        str_ += f"  Has targets configuration: {self.has_targets_configuration()}\n"
        str_ += (
            f"  Has a fitness calculator configuration: "
            f"{self.has_fitness_calculator_configuration()}\n"
        )
        str_ += f"  Has a model configuration: {self.has_model_configuration()}\n\n"

        if pathlib.Path("./checkpoints/").is_dir():
            checkpoints = glob.glob("./checkpoints/**/*.pkl", recursive=True)
            template_path = self.emodel_metadata.as_string()
            checkpoints = [c for c in checkpoints if template_path in c]
            str_ += "OPTIMISATION STATUS\n"
            str_ += f"  Number of checkpoints: {len(checkpoints)}\n"
            for c in checkpoints:
                run, seed = read_checkpoint(c)
                str_ += f"    Seed {seed};"
                str_ += f" Last generation: {run['logbook'].select('gen')[-1]};"
                str_ += f" Best fitness: {sum(run['halloffame'][0].fitness.values)}\n"
            str_ += "\n"

        str_ += "EMODELS BUILDING STATUS\n"
        emodels = self.get_emodels()
        if emodels:
            n_emodels = len(emodels)
            n_emodels_validated = len([e for e in emodels if e.passed_validation])
            idx_best = numpy.argmin([e.fitness for e in emodels])
            str_ += f"  Emodels stored: {n_emodels}\n"
            str_ += (
                f"  Number of validated emodel: {n_emodels_validated} "
                f"({n_emodels_validated / n_emodels} %)\n"
            )
            str_ += (
                f"  Best emodel: seed {emodels[idx_best].seed} with fitness"
                f" {emodels[idx_best].fitness}\n\n"
            )
        else:
            str_ += "  No emodels\n\n"

        return str_



"""Noise stimulus class"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
from importlib import resources

import numpy

from bluepyemodel.ecode.noise import NoiseMixin

logger = logging.getLogger(__name__)


class NoiseOU3(NoiseMixin):
    """NoiseOU3 current stimulus"""

    name = "NoiseOU3"

    def __init__(self, location, **kwargs):
        """Constructor

        Args:
            location(Location): location of stimulus
            **kwargs: See below

        Keyword Arguments:
            holding_current (float): amplitude of the holding current (nA)
            mu (float): mu is a factor in [nA] changing the noise from the file as
                Noise_{injected} = Noise_{file} * (mu/2) + mu
            data_filepath (str): path to the noise .txt data file
                If not given, will use the default one at bluepyemodel/ecodes/data/NoiseOU3.txt
        """

        self.holding_current = kwargs.get("holding_current", None)
        self.threshold_current = None

        self.mu = kwargs.get("mu", None)
        data_filepath = kwargs.get("data_filepath", None)

        if data_filepath is not None:
            series = numpy.loadtxt(data_filepath)
        else:
            resource = resources.files("bluepyemodel").joinpath("ecode", "data", "NoiseOU3.txt")
            with open(resource, "r") as f:
                series = numpy.loadtxt(f)

        self.current_series = series[:, 1]
        self.time_series = series[:, 0]

        super().__init__(
            location=location,
        )



"""HyperDepol stimulus class"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

import numpy

from bluepyemodel.ecode.stimulus import BPEM_stimulus

logger = logging.getLogger(__name__)


class HyperDepol(BPEM_stimulus):
    """HyperDepol current stimulus

    The hyperpolarizing step is usually fixed at 100% of rheobase, and the hyperpolarizing step
    can usually vary from -40% to -160% of rheobase.

    .. code-block:: none

              holdi        holdi+hyper_amp      holdi+depol_amp       holdi
                :                :                     :                :
                :                :           _____________________      :
                :                :          |                     |     :
                :                :          |                     |     :
                :                :          |                     |     :
                :                :          |                     |     :
                :                :          |                     |     :
        |_______________         :          |                     |___________
        ^               |        :          |                     ^           ^
        :               |___________________|                     :           :
        :               ^                   ^                     :           :
        :               :                   :                     :           :
        :               :                   :                     :           :
        t=0             delay               tmid                  toff        totduration
    """

    name = "HyperDepol"

    def __init__(self, location, **kwargs):
        """Constructor
        Args:
            location(Location): location of stimulus
        """

        amp = kwargs.get("amp", None)
        amp2 = kwargs.get("amp2", None)
        hyper_amp_rel = kwargs.get("thresh_perc", None)
        self.hyper_amp = kwargs.get("hyper_amp", None)
        self.hyper_amp_rel = kwargs.get("hyper_amp_rel", None)
        self.depol_amp = kwargs.get("depol_amp", None)
        self.depol_amp_rel = kwargs.get("depol_amp_rel", None)

        if self.hyper_amp is None and amp is not None:
            self.hyper_amp = amp
        if self.depol_amp is None and amp2 is not None:
            self.depol_amp = amp2
        if self.hyper_amp_rel is None and hyper_amp_rel is not None:
            self.hyper_amp_rel = hyper_amp_rel

        if self.hyper_amp is None and self.hyper_amp_rel is None:
            raise TypeError(
                f"In stimulus {self.name}, hyper_amp and hyper_amp_rel cannot be both None."
            )

        if self.depol_amp is None and self.depol_amp_rel is None:
            raise TypeError(
                f"In stimulus {self.name}, depol_amp and depol_amp_rel cannot be both None."
            )

        self.holding_current = kwargs.get("holding_current", None)
        self.threshold_current = None

        self.delay = kwargs.get("delay", 250.0)
        self.tmid = kwargs.get("tmid", 700.0)
        self.toff = kwargs.get("toff", 970.0)
        self.total_duration = kwargs.get("totduration", 1220.0)

        super().__init__(
            location=location,
        )

    @property
    def stim_start(self):
        return self.delay

    @property
    def stim_end(self):
        return self.toff

    @property
    def amplitude(self):
        if self.hyper_amp_rel is None or self.threshold_current is None:
            return self.hyper_amp
        return self.threshold_current * (float(self.hyper_amp_rel) / 100.0)

    @property
    def depol_amplitude(self):
        if self.depol_amp_rel is None or self.threshold_current is None:
            return self.depol_amp
        return self.threshold_current * (float(self.depol_amp_rel) / 100.0)

    def instantiate(self, sim=None, icell=None):
        """Run stimulus"""

        holding_current = self.holding_current if self.holding_current is not None else 0

        icomp = self.location.instantiate(sim=sim, icell=icell)

        self.iclamp = sim.neuron.h.IClamp(icomp.x, sec=icomp.sec)
        self.iclamp.dur = self.total_duration

        self.current_vec = sim.neuron.h.Vector()
        self.time_vec = sim.neuron.h.Vector()

        self.time_vec.append(0.0)
        self.current_vec.append(holding_current)

        self.time_vec.append(self.delay)
        self.current_vec.append(holding_current)

        self.time_vec.append(self.delay)
        self.current_vec.append(holding_current + self.amplitude)

        self.time_vec.append(self.tmid)
        self.current_vec.append(holding_current + self.amplitude)

        self.time_vec.append(self.tmid)
        self.current_vec.append(holding_current + self.depol_amplitude)

        self.time_vec.append(self.toff)
        self.current_vec.append(holding_current + self.depol_amplitude)

        self.time_vec.append(self.toff)
        self.current_vec.append(holding_current)

        self.time_vec.append(self.total_duration)
        self.current_vec.append(holding_current)

        self.iclamp.delay = 0
        self.current_vec.play(
            self.iclamp._ref_amp,  # pylint:disable=W0212
            self.time_vec,
            1,
            sec=icomp.sec,
        )

    def generate(self, dt=0.1):
        """Return current time series

        WARNING: do not offset ! This is on-top of a holding stimulus."""
        holding_current = self.holding_current if self.holding_current is not None else 0
        t = numpy.arange(0.0, self.total_duration, dt)
        current = numpy.full(t.shape, holding_current, dtype="float64")

        ton = int(self.delay / dt)
        tmid = int(self.tmid / dt)
        toff = int(self.toff / dt)

        current[ton:tmid] += self.amplitude
        current[tmid:toff] += self.depol_amplitude

        return t, current



"""Export the emodels in the SONATA format"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import pathlib
import shutil

logger = logging.getLogger(__name__)


def get_output_path_from_metadata(output_base_dir, emodel_metadata, seed, use_allen_notation=True):
    """Get the output path from the emodel_metadata.

    Args:
        output_base_dir (str): output base directory
        emodel_metadata (EModelMetadata): emodel metadata
        seed (int): seed
    """
    return (
        f"./{output_base_dir}/"
        f"{emodel_metadata.as_string(seed=seed, use_allen_notation=use_allen_notation)}/"
    )


def get_output_path(
    emodel,
    output_dir=None,
    output_base_dir="export_emodels_hoc",
    use_allen_notation=True,
    create_dir=True,
):
    """Get the output path.

    Args:
        emodel (EModel): emodel
        output_dir (str): output directory
        output_base_dir (str): if output_dir is None, export to this directory instead,
            using also emodel metadata in the path
        use_allen_notation (bool): whether to replace brain region by its Allen notation
        create_dir (bool): whether to create the output folder if not existent
    """
    if output_dir is None:
        output_dir = get_output_path_from_metadata(
            output_base_dir, emodel.emodel_metadata, emodel.seed, use_allen_notation
        )
    output_path = pathlib.Path(output_dir)
    if create_dir:
        output_path.mkdir(parents=True, exist_ok=True)

    return output_path


def get_hoc_file_path(output_path):
    """Get the hoc file path."""
    output_path = pathlib.Path(output_path)
    return str(output_path / "model.hoc")


def copy_hocs_to_new_output_path(emodel, output_base_dir):
    """Copy the hocs from the local output path to the new nexus output path."""
    old_output_path = get_output_path(
        emodel,
        output_dir=None,
        output_base_dir=output_base_dir,
        use_allen_notation=False,
        create_dir=False,
    )
    output_path_allen = get_output_path(
        emodel,
        output_dir=None,
        output_base_dir=output_base_dir,
        use_allen_notation=True,
        create_dir=False,
    )

    if (
        not pathlib.Path(get_hoc_file_path(output_path_allen)).is_file()
        and pathlib.Path(get_hoc_file_path(old_output_path)).is_file()
    ):
        shutil.copytree(old_output_path, output_path_allen)


def select_emodels(
    emodel_name,
    emodels,
    only_validated=False,
    only_best=True,
    seeds=None,
    iteration=None,
):
    if not emodels:
        logger.warning("In export_emodels_nexus, no emodel for %s", emodel_name)
        return []

    if iteration:
        emodels = [model for model in emodels if model.emodel_metadata.iteration == iteration]

    if only_best:
        emodels = [sorted(emodels, key=lambda x: x.fitness)[0]]

    if seeds:
        emodels = [e for e in emodels if e.seed in seeds]
        if not emodels:
            logger.warning(
                "In export_emodels_nexus, no emodel for %s and seeds %s",
                emodel_name,
                seeds,
            )
            return []

    if only_validated:
        emodels = [e for e in emodels if e.passed_validation]
        if not emodels:
            logger.warning(
                "In export_emodels_nexus, no emodel for %s that passed validation",
                emodel_name,
            )
            return []

    return emodels
